from abc import abstractmethod
from typing import Sequence, Union

import bmesh
import bpy
import numpy as np

from .base import RenderableObject
from ..colors import UniformColors
from ..colors.base import ColorsList, Colors
from ..internal.types import Vector3d
from ..materials.base import Material, MaterialList


# =================================================== Mesh Primitives ==================================================
class MeshPrimitive(RenderableObject):
    """Base class for mesh primitives. Used to throw Exceptions for non-implemented Colors
    subclasses (only UniformColors is supported) and add shared method for setting smooth shading.

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene.
    """
    @abstractmethod
    def __init__(
        self,
            faces_material: Sequence[Sequence[int]] = None,
        **kwargs
    ):
        """Passes all arguments to the constructor of the base class

        Args:
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            blender_object (bpy.types.Object): instance of Blender Object that is wrapped by the class
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        self._faces_material = faces_material
        super().__init__(**kwargs)


    def _blender_set_colors(self, colors: Union[Colors, ColorsList]):
        colors_list = [colors] if isinstance(colors, Colors) else colors
        if not all(isinstance(x, UniformColors) for x in colors_list):
            raise NotImplementedError("Non-uniform colors or textures are not supported in primitives, "
                                      "consider creating a primitive through Mesh for that")
        super()._blender_set_colors(colors_list)

    def set_smooth(self, smooth: bool = True):
        """Enables or disables the smooth surface imitation for the object

        Args:
            smooth (bool): Whether to turn the smooth surface on or off
        """
        bpy.context.view_layer.objects.active = self._blender_object
        bpy.ops.object.mode_set(mode='OBJECT')
        if smooth:
            bpy.ops.object.shade_smooth()
        else:
            bpy.ops.object.shade_flat()

    def _blender_assign_materials(self):
        super()._blender_assign_materials()
        if not (len(self._material_instances) == 1 or self._faces_material is None):
            bpy.context.view_layer.objects.active = self._blender_object
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_mode(type="FACE")
            bpy.ops.mesh.select_all(action='DESELECT')
            bm = bmesh.from_edit_mesh(self._blender_mesh)
            assert self._faces_material is None or (len(self._faces_material) == len(bm.faces)), \
                f"Number of material faces should be equal to the number of faces ({len(bm.faces)})"
            for mat_ind in range(self._materials_count):
                for face in bm.faces:
                    perface_mat_ind = self._faces_material[face.index]
                    if perface_mat_ind == mat_ind:
                        face.select = True
                self._blender_object.active_material_index = mat_ind
                bpy.ops.object.material_slot_assign()
                bpy.ops.mesh.select_all(action='DESELECT')
            # bmesh.update_edit_mesh(self._blender_object.data)
            bpy.ops.object.mode_set(mode='OBJECT')


class CubeMesh(MeshPrimitive):
    """Cube mesh primitive, supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene.

    Methods:
        set_smooth(bool): turns smooth shading on and off based on the bool argument.
    """
    def __init__(
        self,
        size: float,
        tag: str,
        **kwargs
    ):
        """Creates Blender Object that represent Cube mesh primitive

        Args:
            size (float): size of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        obj = self._blender_create_object(size, tag)
        self._blender_mesh = obj.data
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
        self,
        size: float,
        tag: str
    ):
        bpy.ops.mesh.primitive_cube_add(size=size)
        obj = bpy.context.object
        obj.name = tag
        return obj

    def _blender_set_colors(
        self,
        colors_list: ColorsList
    ):
        """Remembers current color properties, builds a color node for material, sets color information to mesh

        Args:
            colors_list (ColorsList): list of target colors
        """
        bpy.context.view_layer.objects.active = self._blender_object
        bpy.ops.object.shade_smooth()
        # bpy.context.space_data.context = 'MODIFIER'
        bpy.ops.object.modifier_add(type='EDGE_SPLIT')
        super()._blender_set_colors(colors_list)


class CircleMesh(MeshPrimitive):
    """Circle mesh primitive, supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene

    Methods:
        set_smooth(bool): turns smooth shading on and off based on the bool argument.
    """
    def __init__(
        self,
        radius: float,
        tag: str,
        num_vertices: int = 32,
        fill_type: str = "NGON",
        **kwargs
    ):
        """Creates Blender Object that represent Circle mesh primitive

        Args:
            radius (float): radius of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            num_vertices (int, optional): number of vertices in primitive in [3, 10000000] (default: 32)
            fill_type (str, optional): fill type, one of [NOTHING, NGON, TRIFAN] (default: NGON)
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        obj = self._blender_create_object(num_vertices, radius, fill_type, tag)
        self._blender_mesh = obj.data
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
        self,
        num_vertices: int,
        radius: float,
        fill_type: str,
        tag: str
    ):
        bpy.ops.mesh.primitive_circle_add(vertices=num_vertices, radius=radius, fill_type=fill_type)
        obj = bpy.context.object
        obj.name = tag
        return obj

    def _blender_set_colors(
        self,
        colors_list: Colors
    ):
        """Remembers current color properties, builds a color node for material, sets color information to mesh

        Args:
            colors_list (ColorsList): list of target colors
        """
        bpy.context.view_layer.objects.active = self._blender_object
        bpy.ops.object.shade_smooth()
        # bpy.context.space_data.context = 'MODIFIER'
        bpy.ops.object.modifier_add(type='EDGE_SPLIT')
        super()._blender_set_colors(colors_list)


class CylinderMesh(MeshPrimitive):
    """Cylinder mesh primitive, supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene.

    Methods:
        set_smooth(bool): turns smooth shading on and off based on the bool argument.
    """
    def __init__(
        self,
        radius: float,
        height: float,
        tag: str,
        num_vertices: int = 32,
        fill_type: str = "NGON",
        **kwargs
    ):
        """Creates Blender Object that represent Cylinder mesh primitive

        Args:
            radius (float): radius of a primitive in [0, inf]
            height (float): height of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            num_vertices (int, optional): number of vertices in primitive in [3, 10000000] (default: 32)
            fill_type (str, optional): fill type, one of [NOTHING, NGON, TRIFAN] (default: NGON)
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        obj = self._blender_create_object(num_vertices, radius, height, fill_type, tag)
        self._blender_mesh = obj.data
        # obj.scale[2] = height / radius
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
        self,
        num_vertices: int,
        radius: float,
        height: float,
        fill_type: str,
        tag: str
    ):
        bpy.ops.mesh.primitive_cylinder_add(vertices=num_vertices, radius=radius, depth=height, end_fill_type=fill_type)
        obj = bpy.context.object
        obj.name = tag
        return obj

    def _blender_set_colors(
            self,
            colors_list: ColorsList
    ):
        """Remembers current color properties, builds a color node for material, sets color information to mesh

        Args:
            colors_list (ColorsList): list of target colors
        """
        bpy.context.view_layer.objects.active = self._blender_object
        bpy.ops.object.shade_smooth()
        # bpy.context.space_data.context = 'MODIFIER'
        bpy.ops.object.modifier_add(type='EDGE_SPLIT')
        super()._blender_set_colors(colors_list)


class PlaneMesh(MeshPrimitive):
    """Plane mesh primitive, supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene.
        shadow_catcher (bool, optional): control whether the object will act as a shadow catcher (i.e. object
        geometry is hidden from the render, only casted shadows are rendered).

    Methods:
        set_smooth(bool): turns smooth shading on and off based on the bool argument.
    """

    def __init__(
            self,
            size: float,
            tag: str,
            shadow_catcher: bool = False,
            **kwargs
    ):
        """Creates Blender Object that represent Plane mesh primitive

        Args:
            size (float): size of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            shadow_catcher (bool, optional): control whether the object will act as a shadow catcher
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        obj = self._blender_create_object(size, tag, shadow_catcher)
        self._blender_mesh = obj.data
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
            self,
            size: float,
            tag: str,
            shadow_catcher: bool
    ):
        bpy.ops.mesh.primitive_plane_add(size=size)
        obj = bpy.context.object
        obj.name = tag

        if shadow_catcher:
            obj.is_shadow_catcher = True
            obj.visible_glossy = False
            obj.visible_diffuse = False
            obj.visible_transmission = False
            obj.visible_volume_scatter = False

        return obj

    def _blender_set_colors(
            self,
            colors_list: ColorsList
    ):
        """Remembers current color properties, builds a color node for material, sets color information to mesh

        Args:
            colors_list (ColorsList): list of target colors
        """
        bpy.context.view_layer.objects.active = self._blender_object
        bpy.ops.object.shade_smooth()
        # bpy.context.space_data.context = 'MODIFIER'
        bpy.ops.object.modifier_add(type='EDGE_SPLIT')
        super()._blender_set_colors(colors_list)
# =============================================== End of Mesh Primitives ===============================================


# ================================================ Parametric Primitives ===============================================
class ParametricPrimitive(RenderableObject):
    """Base class for parametric primitives. Used to throw Exceptions for non-implemented Colors
    subclasses (only UniformColors is supported) and non-implemented Materials (only one Material is supported).

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene.
    """

    @abstractmethod
    def __init__(
        self,
        **kwargs
    ):
        """Passes all arguments to the constructor of the base class

        Args:
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            blender_object (bpy.types.Object): instance of Blender Object that is wrapped by the class
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)

    def _blender_set_materials(self, material_list: MaterialList):
        if not len(material_list) == 1:
            raise NotImplementedError("Multiple materials are not supported in parametric primitives, "
                                      "consider creating a primitive through Mesh for that")
        super()._blender_set_materials(material_list)

    def _blender_set_colors(self, colors_list: ColorsList):
        if  not len(colors_list) == 1 or not isinstance(colors_list[0], UniformColors):
            raise NotImplementedError("Multiple materials, non-uniform colors or textures are not supported "
                                      "in parametric primitives, consider creating a primitive through Mesh for that")
        super()._blender_set_colors(colors_list)


class EllipsoidNURBS(ParametricPrimitive):
    """NURBS Ellipsoid, implemented as NURBS Sphere that is rescaled along axes,
     supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene
    """
    def __init__(
        self,
        radius: Vector3d,
        tag: str,
        **kwargs
    ):
        """Creates Blender Object that represent NURBS Surface Sphere primitive with different scales along axis,
        resulting in Ellipsoid shape

        Args:
            radius (float): radius of a primitive in [0, inf]
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        obj = self._blender_create_object(radius, tag)
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
            self,
            radius: Vector3d,
            tag: str
    ):
        bpy.ops.surface.primitive_nurbs_surface_sphere_add(radius=radius[0])
        obj = bpy.context.object
        obj.scale = (np.array(radius) / radius[0]).tolist()
        obj.name = tag
        return obj


class SphereNURBS(EllipsoidNURBS):
    """NURBS Sphere, supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene
    """
    def __init__(
        self,
        radius: float,
        tag: str,
        **kwargs
    ):
        """Creates Blender Object that represent NURBS Surface Sphere primitive

        Args:
            radius (float): radius of a primitive in [0, inf]
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs, radius=(radius, radius, radius), tag=tag)


class CurveBezier(ParametricPrimitive):
    """Bezier Curve, supports only uniform coloring (UniformColors)

    Properties:
        emit_shadow (bool, optional): control whether the object will emit shadow from any light source in the scene
    """
    def __init__(
        self,
        keypoints: np.ndarray,
        radius: float,
        tag: str,
        **kwargs
    ):
        """Creates Blender Object that represent Bezier Curve primitive - a tube passing through the given keypoints

        Args:
            keypoints (np.ndarray): keypoints for the curve
            radius (float): radius of a tube in [0, inf]
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        obj = self._blender_create_object(tag)
        obj.data.bevel_depth = radius
        obj.data.bevel_resolution = 4
        if len(keypoints) > 2:
            obj.data.splines[0].bezier_points.add(count=len(keypoints) - 2)
        for ind, coords in enumerate(keypoints):
            obj.data.splines[0].bezier_points[ind].co = coords
            obj.data.splines[0].bezier_points[ind].handle_left_type = 'VECTOR'
            obj.data.splines[0].bezier_points[ind].handle_right_type = 'VECTOR'
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
        self,
        tag: str
    ):
        bpy.ops.curve.primitive_bezier_curve_add()
        obj = bpy.context.object
        obj.data.dimensions = '3D'
        obj.data.fill_mode = 'FULL'
        return obj
# ============================================ End of Parametric Primitives ============================================



import math
from collections import defaultdict

import bpy
import numpy as np
from scipy.spatial.transform import Rotation


def move_collection(parent, collection):
    # from https://blender.stackexchange.com/questions/157828/how-to-duplicate-a-certain-collection-using-python
    def _move_collection(parent, collection):
        # re-link objects
        for o in collection.objects:
            collection.objects.unlink(o)
            parent.objects.link(o)

        for c in collection.children:
            # Create child and link it
            cc = bpy.data.collections.new(c.name)
            parent.children.link(cc)
            # Fix naming: blender appends .001, .002, etc. to new names
            orig_name = c.name
            created_name = cc.name
            c.name = created_name
            cc.name = orig_name
            # Recursively move everything that is inside
            _move_collection(cc, c)

        for c in collection.children:
            bpy.data.collections.remove(c, do_unlink=True)

    _move_collection(parent, collection)


def parse_camera_from_blendfile(obj: bpy.types.Object, resolution: np.ndarray):
    """Parse camera parameters from blender object.
    Args:
        obj (bpy.types.Object): Blender object with camera data
        resolution (np.ndarray): array of [resolution_x, resolution_y]

    Returns:
        camera_type (str): ORTHO, PRESP or NONE
        camera_dict (dict): dictionary with camera parameters
    """
    camera_dict = dict()
    camera_dict["tag"] = obj.name
    camera_dict["resolution"] = resolution

    # position
    camera_dict["translation"] = np.array(obj.location)
    if obj.rotation_mode == "QUATERNION":
        camera_dict["rotation"] = np.array(obj.rotation_quaternion)
    elif obj.rotation_mode == "AXIS_ANGLE":
        rotvec = np.array(obj.rotation_axis_angle)
        angle, axis = rotvec[0], rotvec[1:]
        rotvec = (axis / np.linalg.norm(axis)) * angle

        rot = Rotation.from_rotvec(rotvec)
        camera_dict["rotation"] = np.roll(rot.as_quat(), 1)
    else:
        # euler angles
        rot_data = obj.rotation_euler
        mode = rot_data.order
        angles = np.array(rot_data[:])

        rot = Rotation.from_euler(mode.lower(), angles, degrees=False)
        camera_dict["rotation"] = np.roll(rot.as_quat(), 1)

    # camera parameters
    if obj.data.type == "ORTHO":
        # scale
        camera_dict["ortho_scale"] = obj.data.ortho_scale

        # near / far
        camera_dict["near"] = obj.data.clip_start
        camera_dict["far"] = obj.data.clip_end
    elif obj.data.type == "PERSP":
        # sensor params
        fov_x, fov_y = None, None
        sensor_fit = obj.data.sensor_fit

        # determine sensor fit
        if sensor_fit == "AUTO":
            sensor_fit = "HORIZONTAL" if resolution[0] >= resolution[1] else "VERTICAL"

        # determine fov
        if sensor_fit == "HORIZONTAL":
            if obj.data.lens_unit == "MILLETERS":
                fov_x = 2 * math.atan(0.5 * obj.data.sensor_width / obj.data.lens)
            else:
                fov_x = obj.data.angle_x
        else:  # VERTICAL
            if obj.data.lens_unit == "MILLETERS":
                fov_y = 2 * math.atan(0.5 * obj.data.sensor_height / obj.data.lens)
            else:
                fov_y = obj.data.angle_y
        camera_dict["fov_x"] = fov_x
        camera_dict["fov_y"] = fov_y

        # center
        ideal_center = resolution / 2.
        center_offset = np.array([obj.data.shift_x, obj.data.shift_y])
        camera_dict["center"] = ideal_center / resolution - center_offset

        # near / far
        camera_dict["near"] = obj.data.clip_start
        camera_dict["far"] = obj.data.clip_end

    return obj.data.type, camera_dict

def parse_light_from_blendfile(obj: bpy.types.Object):
    # Parse common parameters
    translation, quaternion, scale = obj.matrix_world.decompose()
    # "scale", "multiple_importance", "shadow_caustics" are not supported
    light_dict = {
        "strength": obj.data.energy,
        "color": np.array(obj.data.color),
        "cast_shadows": obj.data.cycles.cast_shadow,
        "rotation": quaternion,
        "translation": translation,
        "tag": obj.name
    }

    # Parse type specific parameters
    light_type = obj.data.type
    if light_type == "POINT":
        light_dict["shadow_soft_size"] = obj.data.shadow_soft_size
    elif light_type == "SUN":
        light_dict["angular_diameter"] = obj.data.angle
    elif light_type == "SPOT":
        light_dict["spot_size"] = obj.data.spot_size
        light_dict["spot_blend"] = obj.data.spot_blend
        light_dict["shadow_soft_size"] = obj.data.shadow_soft_size
    elif light_type == "AREA":
        light_dict["shape"] = obj.data.shape.lower()
        if light_dict["shape"] in ["RECTANGLE", "ELLIPSE"]:
            light_dict["size"] = np.array([obj.data.size, obj.data.size_y], dtype=np.float32)
        else:
            light_dict["size"] = obj.data.size

    return light_type, light_dict


import copy
import functools
import warnings

import cv2
import numba
import numpy as np


class Camera:
    def __init__(
            self, optical_center=None, rot_world_to_cam=None, intrinsic_matrix=np.eye(3),
            distortion_coeffs=None, world_up=(0, 0, 1), extrinsic_matrix=None,
            trans_after_rot=None, dtype=np.float32):
        """Pinhole camera with extrinsic and intrinsic calibration with optional distortions.

        The camera coordinate system has the following axes:
          x points to the right
          y points down
          z points forwards

        The world z direction is assumed to point up by default, but `world_up` can also be
         specified differently.

        Args:
            optical_center: position of the camera in world coordinates (eye point)
            rot_world_to_cam: 3x3 rotation matrix for transforming column vectors
                from being expressed in world reference frame to being expressed in camera
                reference frame as follows:
                column_point_cam = rot_matrix_world_to_cam @ (column_point_world - optical_center)
            intrinsic_matrix: 3x3 matrix that maps 3D points in camera space to homogeneous
                coordinates in image (pixel) space. Its last row must be (0,0,1).
            distortion_coeffs: parameters describing radial and tangential lens distortions,
                following OpenCV's model and order: k1, k2, p1, p2, k3 or None,
                if the camera has no distortion.
            world_up: a world vector that is designated as "pointing up".
            extrinsic_matrix: 4x4 extrinsic transformation matrix as an alternative to
                providing `optical_center` and `rot_world_to_cam`.
            trans_after_rot: translation vector to apply after the rotation
                (alternative to optical_center, which is a negative translation before the rotation)
        """

        if optical_center is not None and extrinsic_matrix is not None:
            raise Exception(
                'Provide only one of `optical_center`, `trans_after_rot` or `extrinsic_matrix`!')
        if extrinsic_matrix is not None and rot_world_to_cam is not None:
            raise Exception('Provide only one of `rot_world_to_cam` or `extrinsic_matrix`!')

        if (optical_center is None) and (trans_after_rot is None) and (extrinsic_matrix is None):
            optical_center = np.zeros(3, dtype=dtype)

        if (rot_world_to_cam is None) and (extrinsic_matrix is None):
            rot_world_to_cam = np.eye(3, dtype=dtype)

        if extrinsic_matrix is not None:
            self.R = np.asarray(extrinsic_matrix[:3, :3], dtype=dtype)
            self.t = -self.R.T @ extrinsic_matrix[:3, 3].astype(dtype)
        else:
            self.R = np.asarray(rot_world_to_cam, dtype=dtype)
            if optical_center is not None:
                self.t = np.asarray(optical_center, dtype=dtype)
            else:
                self.t = -self.R.T @ np.asarray(trans_after_rot, dtype=dtype)

        self.intrinsic_matrix = np.asarray(intrinsic_matrix, dtype=dtype)
        if distortion_coeffs is None:
            self.distortion_coeffs = None
        else:
            self.distortion_coeffs = np.asarray(distortion_coeffs, dtype=dtype)
            if np.all(self.distortion_coeffs == 0):
                self.distortion_coeffs = None

        self.world_up = np.asarray(world_up, dtype=dtype)

        if not np.allclose(self.intrinsic_matrix[2, :], [0, 0, 1]):
            raise Exception(f'Bottom row of intrinsic matrix must be (0,0,1), '
                            f'got {self.intrinsic_matrix[2, :]}.')

    # Methods to transform between coordinate systems (world, camera, image)
    def camera_to_image(self, points):
        """Transform points from 3D camera coordinate space to image space."""

        if self.has_distortion():
            with warnings.catch_warnings():
                warnings.simplefilter(
                    'ignore', category=numba.errors.NumbaPerformanceWarning)
                return project_points(points, self.distortion_coeffs, self.intrinsic_matrix)
        else:
            projected = points[..., :2] / points[..., 2:]
            return projected @ self.intrinsic_matrix[:2, :2].T + self.intrinsic_matrix[:2, 2]

    def world_to_camera(self, points):
        return (points - self.t) @ self.R.T

    def camera_to_world(self, points):
        return points @ self.R + self.t

    def world_to_image(self, points):
        return self.camera_to_image(self.world_to_camera(points))

    def image_to_camera(self, points, depth=1):
        if not self.has_distortion():
            normalized_points = (
                ((points - self.intrinsic_matrix[:2, 2]) @
                 np.linalg.inv(self.intrinsic_matrix[:2, :2])))
            return cv2.convertPointsToHomogeneous(normalized_points)[:, 0, :] * depth

        points = np.expand_dims(np.asarray(points, np.float32), 0)

        if len(self.distortion_coeffs) == 4:
            new_image_points = cv2.fisheye.undistortPoints(
                points, self.intrinsic_matrix, self.distortion_coeffs, None, None, None)
            # To match the behavior of the 5-parameter undistortPoints, we transpose the result
            new_image_points = np.transpose(new_image_points, [1, 0, 2])
        else:
            new_image_points = cv2.undistortPoints(
                points, self.intrinsic_matrix, self.distortion_coeffs, None, None, None)
        return cv2.convertPointsToHomogeneous(new_image_points)[:, 0, :] * depth

    def image_to_world(self, points, camera_depth=1):
        return self.camera_to_world(self.image_to_camera(points, camera_depth))

    def scale_output(self, factor):
        """Adjusts the camera such that the images become scaled by `factor`. It's a scaling with
        the origin as anchor point.
        The difference with `self.zoom` is that this method also moves the principal point,
        multiplying its coordinates by `factor`."""
        self.intrinsic_matrix[:2] *= np.expand_dims(np.float32(factor), -1)

    def undistort(self, alpha_balance=None, new_imshape=None):
        if alpha_balance is not None and self.has_distortion():
            imsize = tuple(imshape[:2][::-1])
            new_imsize = imsize if new_imshape is None else tuple(new_imshape[:2][::-1])
            if len(self.distortion_coeffs) == 4:
                self.intrinsic_matrix = cv2.fisheye.estimateNewCameraMatrixForUndistortRectify(
                    self.intrinsic_matrix, self.distortion_coeffs, imsize, np.eye(3),
                    new_size=new_imsize, balance=alpha_balance)
            else:
                self.intrinsic_matrix = cv2.getOptimalNewCameraMatrix(
                    self.intrinsic_matrix, self.distortion_coeffs, imsize, alpha_balance,
                    new_imsize, centerPrincipalPoint=False)[0]

        self.distortion_coeffs = None

    def square_pixels(self):
        """Adjusts the intrinsic matrix such that the pixels correspond to squares on the
        image plane."""
        fx = self.intrinsic_matrix[0, 0]
        fy = self.intrinsic_matrix[1, 1]
        fmean = 0.5 * (fx + fy)
        multiplier = np.array([[fmean / fx, 0, 0], [0, fmean / fy, 0], [0, 0, 1]], np.float32)
        self.intrinsic_matrix = multiplier @ self.intrinsic_matrix

    def center_principal_point(self, imshape):
        """Adjusts the intrinsic matrix so that the principal point becomes located at the center
        of an image sized imshape (height, width)"""

        self.intrinsic_matrix[:2, 2] = np.float32([(imshape[1] - 1) / 2, (imshape[0] - 1) / 2])

    def get_distortion_coeffs(self):
        if self.distortion_coeffs is None:
            return np.zeros(shape=(5,), dtype=np.float32)
        return self.distortion_coeffs

    def has_distortion(self):
        return self.distortion_coeffs is not None and not np.all(self.distortion_coeffs == 0)

    def copy(self):
        return copy.deepcopy(self)

    @property
    def optical_center(self):
        return self.t


def reproject_image(
        image, old_camera, new_camera, output_imshape, border_mode=cv2.BORDER_CONSTANT,
        border_value=0, interp=None, antialias_factor=1, dst=None):
    """Transform an `image` captured with `old_camera` to look like it was captured by
    `new_camera`. The optical center (3D world position) of the cameras must be the same, otherwise
    we'd have parallax effects and no unambiguous way to construct the output.
    Ignores the issue of aliasing altogether.

    Args:
        image: the input image
        old_camera: the camera that captured `image`
        new_camera: the camera that should capture the newly returned image
        output_imshape: (height, width) for the output image
        border_mode: OpenCV border mode for treating pixels outside `image`
        border_value: OpenCV border value for treating pixels outside `image`
        interp: OpenCV interpolation to be used for resampling.
        antialias_factor: If larger than 1, first render a higher resolution output image
            that is `antialias_factor` times larger than `output_imshape` and subsequently resize
            it by 'area' interpolation to the desired size.
        dst: destination array (optional)

    Returns:
        The new image.
    """
    if antialias_factor == 1:
        return reproject_image_aliased(
            image, old_camera, new_camera, output_imshape, border_mode, border_value, interp,
            dst=dst)

    new_camera = new_camera.copy()
    a = antialias_factor
    new_camera.scale_output(a)
    new_camera.intrinsic_matrix[:2, 2] += (a - 1) / 2
    intermediate_imshape = (a * output_imshape[0], a * output_imshape[1])
    result = reproject_image_aliased(
        image, old_camera, new_camera, intermediate_imshape, border_mode, border_value, interp)
    return cv2.resize(
        result, dsize=(output_imshape[1], output_imshape[0]),
        interpolation=cv2.INTER_AREA, dst=dst)


def reproject_image_aliased(
        image, old_camera, new_camera, output_imshape, border_mode=cv2.BORDER_CONSTANT,
        border_value=0, interp=None, dst=None):
    """Transform an `image` captured with `old_camera` to look like it was captured by
    `new_camera`. The optical center (3D world position) of the cameras must be the same, otherwise
    we'd have parallax effects and no unambiguous way to construct the output.
    Aliasing issues are ignored.
    """

    if interp is None:
        interp = cv2.INTER_LINEAR

    map1, map2 = get_maps_for_remap(old_camera, new_camera, output_imshape)
    remapped = cv2.remap(
        image, map1, map2, interp, borderMode=border_mode, borderValue=border_value, dst=dst)
    if remapped.ndim < image.ndim:
        return np.expand_dims(remapped, -1)
    return remapped


@functools.lru_cache(maxsize=10)
def get_maps_for_remap(old_camera, new_camera, output_imshape):
    new_maps = get_grid_coords((output_imshape[0], output_imshape[1]))
    newim_coords = new_maps.reshape([-1, 2])
    if not new_camera.has_distortion():
        partial_homography = (
                old_camera.R @ np.linalg.inv(new_camera.R) @
                np.linalg.inv(new_camera.intrinsic_matrix))
        new_im_homogeneous = np.squeeze(cv2.convertPointsToHomogeneous(newim_coords), axis=1)
        old_camera_coords = new_im_homogeneous @ partial_homography.T
        oldim_coords = old_camera.camera_to_image(old_camera_coords)
    else:
        world_coords = new_camera.image_to_world(newim_coords)
        oldim_coords = old_camera.world_to_image(world_coords)
    old_maps = oldim_coords.reshape(new_maps.shape).astype(np.float32)
    map1 = old_maps[..., 0]
    map2 = old_maps[..., 1]
    return map1, map2


def allclose_or_nones(a, b):
    """Check if all corresponding values in arrays a and b are close to each other in the sense of
    np.allclose, or both a and b are None, or one is None and the other is filled with zeros.
    """

    if a is None and b is None:
        return True

    if a is None:
        return cv2.countNonZero(b) == 0

    if b is None:
        return cv2.countNonZero(a) == 0

    return np.allclose(a, b)


@numba.njit()
def project_points(points, dist_coeff, intrinsic_matrix):
    intrinsic_matrix = intrinsic_matrix.astype(np.float32)
    points = points.astype(np.float32)
    proj = points[..., :2] / points[..., 2:]

    if dist_coeff.shape[0] == 4:
        # Fisheye distortion
        # https://docs.opencv.org/4.x/db/d58/group__calib3d__fisheye.html
        # coefficients: k1, k2, k3, k4
        r = np.sqrt(np.sum(np.square(proj), axis=1))  # np.linalg.norm(proj, axis=1)
        t = np.arctan(r)
        t2 = np.square(t)
        t_d = (
                ((((dist_coeff[3] * t2 + dist_coeff[2]) * t2 + dist_coeff[1]) * t2 +
                  dist_coeff[0]) * t2 + np.float32(1.0)) * t)
        # with np.errstate(invalid='ignore'):
        proj *= np.nan_to_num(np.expand_dims(t_d / r, 1))
    else:
        # Standard distortion (radial, tangential, thin prism)
        # coefficient order: k1, k2, p1, p2, k3, k4, k5, k6, s1, s2, s3, s4
        r2 = np.sum(np.square(proj), axis=1)
        # dist_coeff = np.pad(dist_coeff, (0, 12 - len(dist_coeff)))
        dist_coeff = np.concatenate(
            (dist_coeff, np.zeros((12 - dist_coeff.shape[0],), dtype=np.float32)),
            axis=0)
        distorter = (
                ((((dist_coeff[4] * r2 + dist_coeff[1]) * r2 + dist_coeff[0]) * r2 + np.float32(
                    1.0)) /
                 (((dist_coeff[7] * r2 + dist_coeff[6]) * r2 + dist_coeff[5]) * r2 + np.float32(
                     1.0))) +
                + np.float32(2.0) * np.sum(proj * dist_coeff[3:1:-1], axis=1))

        proj[:] = (
                proj * np.expand_dims(distorter, 1) +
                (dist_coeff[9:12:2] * np.expand_dims(r2, 1) +
                 dist_coeff[8:11:2] + dist_coeff[3:1:-1]) * np.expand_dims(r2, 1))

    # Scheimpflug extension from opencv is not used
    # For that, one would need to implement tilt distortion
    # proj[:] = tilt_distort(proj, dist_coeff[12:14])
    return (proj @ intrinsic_matrix[:2, :2].T + intrinsic_matrix[:2, 2]).astype(np.float32)


@functools.lru_cache(5)
def get_grid_coords(output_imshape):
    """Return a meshgrid of coordinates for the image shape `output_imshape` (height, width).

    Returns
        Meshgrid of shape [height, width, 2], with the x and y coordinates (in this order)
            along the last dimension. DType float32.
    """
    y, x = np.mgrid[:output_imshape[0], :output_imshape[1]].astype(np.float32)
    return np.stack([x, y], axis=-1)



from typing import Sequence

import bmesh
import bpy
import numpy as np
import itertools
from .base import RenderableObject
from ..colors import VertexColors, UniformColors
from ..colors.base import ColorsList
from ..colors.texture import VertexUV, FacesUV, UVColors
from ..materials.base import Material


class Mesh(RenderableObject):
    """Basic mesh, supports uniform (UniformColors), per-vertex (VertexColors)
    and texture (TextureColors and FileTextureColors) coloring with per-vertex uv (VertexUV) or
    per-face uv maps (FacesUV).

    Properties:
        emit_shadow (bool, optional): control whether mesh will emit shadow from any light source in the scene
    """

    def __init__(
            self,
            vertices: np.ndarray,
            faces: np.ndarray,
            tag: str,
            faces_material: Sequence[Sequence[int]] = None,
            **kwargs
    ):
        """Creates Blender Object that represent given mesh

        Args:
            vertices (np.ndarray): mesh vertices
            faces (np.ndarray): mesh faces
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            quaternion (Vector4d, optional): rotation applied to Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
            faces_material (np.ndarray, optional): for each face, the material index assigned to it
        """
        obj = self._blender_create_object(vertices, faces, tag)
        self._faces_material = faces_material
        self._faces_count = len(faces)
        super().__init__(**kwargs, blender_object=obj, tag=tag)

    def _blender_create_object(
            self,
            vertices: np.ndarray,
            faces: np.ndarray,
            tag: str
    ) -> bpy.types.Object:
        """Creates mesh object in Blender

        Args:
            vertices (np.ndarray): mesh vertices
            faces (np.ndarray): mesh faces

        Returns:
            bpy.types.Object: Blender mesh
        """
        mesh = bpy.data.meshes.new(name=tag)
        mesh.from_pydata(vertices.tolist(), [], faces.tolist())
        obj = bpy.data.objects.new(tag, mesh)
        bpy.context.collection.objects.link(obj)
        self._blender_mesh = mesh
        return obj

    def set_smooth(self, smooth: bool = True):
        """Turns smooth shading on and off based on the bool argument

        Args:
            smooth (bool, optional): If True shade smooth else shade flat (default: True)
        """
        bpy.context.view_layer.objects.active = self._blender_object
        bpy.ops.object.mode_set(mode='EDIT')
        bm = bmesh.from_edit_mesh(self._blender_mesh)
        for face in bm.faces:
            face.smooth = smooth
        bpy.ops.object.mode_set(mode='OBJECT')
        if smooth:
            bpy.ops.object.shade_smooth()
        else:
            bpy.ops.object.shade_flat()

    def _blender_set_colors(
            self,
            colors_list: ColorsList
    ):
        """Remembers current color properties, builds a color node for material, sets color information to mesh

        Args:
            colors_list (ColorsList): list of target colors
        """
        for colors in colors_list:
            if isinstance(colors, VertexColors):
                bpy.context.view_layer.objects.active = self._blender_object
                bpy.ops.object.mode_set(mode='EDIT')
                bm = bmesh.from_edit_mesh(self._blender_mesh)
                color_layer = bm.loops.layers.color.new("color")
                for face in bm.faces:
                    for loop in face.loops:
                        loop[color_layer] = colors.vertex_colors[loop.vert.index]
                bpy.ops.object.mode_set(mode='OBJECT')
                self._blender_mesh.vertex_colors["color"].active_render = True
            elif isinstance(colors, UVColors):
                bpy.context.view_layer.objects.active = self._blender_object
                bpy.ops.object.mode_set(mode='EDIT')
                self._blender_mesh.uv_layers.new(name='NewUVMap')
                bm = bmesh.from_edit_mesh(self._blender_mesh)
                uv_layer = bm.loops.layers.uv.active
                uv_map = colors.uv_map
                if isinstance(uv_map, VertexUV):
                    for face in bm.faces:
                        for loop in face.loops:
                            loop[uv_layer].uv = uv_map.data[loop.vert.index].tolist()
                elif isinstance(uv_map, FacesUV):
                    for face in bm.faces:
                        face_uv = uv_map.data[face.index]
                        for loop, loop_uv_coords in zip(face.loops, face_uv):
                            loop[uv_layer].uv = loop_uv_coords.tolist()
                else:
                    raise NotImplementedError(f"Unknown UV map type: {uv_map.__class__.__name__}")
                bpy.ops.object.mode_set(mode='OBJECT')
            elif not isinstance(colors, UniformColors):
                raise NotImplementedError(f"Unknown Colors type {colors.__class__.__name__}")
        super()._blender_set_colors(colors_list)


    def _blender_clear_colors(self):
        """Clears Blender color node and erases node constructor
        """
        for uvl in self._blender_mesh.uv_layers.values():
            self._blender_mesh.uv_layers.remove(uvl)
        super()._blender_clear_colors()


    def update_vertices(
            self,
            vertices: np.ndarray
    ):
        """Updates mesh vertices coordinates

        Args:
            vertices (np.ndarray): new vertex coordinates
        """
        assert len(self._blender_mesh.vertices) == len(vertices), \
            f"Number of vertices should be the same (expected {len(self._blender_mesh.vertices)}, got {len(vertices)})"
        self._blender_mesh.vertices.foreach_set("co", vertices.ravel())
        self._blender_mesh.update()

    def _blender_assign_materials(self):
        super()._blender_assign_materials()
        assert self._faces_material is None or (len(self._faces_material) == self._faces_count), \
            f"Number of material faces should be equal to the number of faces ({self._faces_count})"
        if not (len(self._material_instances) == 1 or self._faces_material is None):
            bpy.context.view_layer.objects.active = self._blender_object
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_mode(type='FACE')
            bpy.ops.mesh.select_all(action='DESELECT')
            bm = bmesh.from_edit_mesh(self._blender_mesh)
            for mat_ind in range(self._materials_count):
                for face in bm.faces:
                    perface_mat_ind = self._faces_material[face.index]
                    if perface_mat_ind == mat_ind:
                        face.select = True
                self._blender_object.active_material_index = mat_ind
                bpy.ops.object.material_slot_assign()
                bpy.ops.mesh.select_all(action='DESELECT')
            # bmesh.update_edit_mesh(self._blender_object.data)
            bpy.ops.object.mode_set(mode='OBJECT')




from typing import Tuple

import bpy

from .base import Material, material_property, MaterialInstance
from .wireframe import WireframeMaterial


class PrincipledBSDFMaterial(Material):
    """A class which manages the parameters of PrincipledBSDF Blender material.
    Full docs: https://docs.blender.org/manual/en/latest/render/shader_nodes/shader/principled.html
    """

    def __init__(
            self, metallic=0.0, specular=0.3, specular_tint=0.0, roughness=0.4, anisotropic=0.0,
            anisotropic_rotation=0.0, sheen=0.0, sheen_tint=0.5, clearcoat=0.0, clearcoat_roughness=0.0,
            ior=1.45, transmission=0.0, transmission_roughness=0.0, emission=(0, 0, 0, 0),
            emission_strength=0.0, alpha=1.0, base_color=(0,0,0,1), use_backface_culling=True, use_colors_for_emission=False
    ):
        super().__init__(use_backface_culling=use_backface_culling)
        self._property2blender_mapping = {
            "metallic": "Metallic", "specular": "Specular", "specular_tint": "Specular Tint", "roughness": "Roughness",
            "anisotropic": "Anisotropic", "anisotropic_rotation": "Anisotropic Rotation", "sheen": "Sheen",
            "sheen_tint": "Sheen Tint", "clearcoat": "Clearcoat", "clearcoat_roughness": "Clearcoat Roughness",
            "ior": "IOR", "transmission": "Transmission", "transmission_roughness": "Transmission Roughness",
            "emission": "Emission", "emission_strength": "Emission Strength", "alpha": "Alpha", "base_color": "Base Color"
        }
        for argname, argvalue in locals().items():
            if argname in self._property2blender_mapping.keys():
                self.__setattr__(argname, material_property(argname))
                self.__setattr__("_" + argname, argvalue)

        self._use_colors_for_emission = use_colors_for_emission

    def create_material(self, name: str = "object_material") -> MaterialInstance:
        """Create the Blender material with the parameters stored in the current object

        Args:
            name (str): a unique material name for Blender

        Returns:
            Tuple[bpy.types.Material, bpy.types.ShaderNodeBsdfPrincipled]: Blender material and the
                shader node which uses the created material
        """

        object_material = bpy.data.materials.new(name=name)
        object_material.use_nodes = True
        object_material.use_backface_culling = self._use_backface_culling
        bsdf_node = object_material.node_tree.nodes["Principled BSDF"]

        color_inp_key = "Emission" if self._use_colors_for_emission else "Base Color"
        material_instance = MaterialInstance(blender_material=object_material,
                                             inputs={"Color": bsdf_node.inputs[color_inp_key], "Alpha": bsdf_node.inputs["Alpha"],
                                                     "Emission": bsdf_node.inputs["Emission"],
                                                     "Emission Strength": bsdf_node.inputs["Emission Strength"]})

        # Set material properties
        for property_name, blender_name in self._property2blender_mapping.items():
            bsdf_node.inputs[blender_name].default_value = self.__getattribute__("_" + property_name)

        return material_instance


class GlossyBSDFMaterial(Material):
    """A class which manages the parameters of GlossyBSDF Blender material.
    Full docs: https://docs.blender.org/manual/en/latest/render/shader_nodes/shader/glossy.html
    """

    def __init__(self, roughness=0.4, distribution="GGX", use_backface_culling=True):
        super().__init__(use_backface_culling=use_backface_culling)
        self.roughness, self._roughness = material_property("roughness"), roughness
        self._distribution = distribution

    def create_material(self, name: str = "object_material") -> MaterialInstance:
        """Create the Blender material with the parameters stored in the current object

        Args:
            name (str): a unique material name for Blender

        Returns:
            Tuple[bpy.types.Material, bpy.types.ShaderNodeBsdfGlossy]: Blender material and the
                shader node which uses the created material
        """

        object_material = bpy.data.materials.new(name=name)
        object_material.use_nodes = True
        object_material.use_backface_culling = self._use_backface_culling
        material_nodes = object_material.node_tree.nodes

        bsdf_node = material_nodes.new("ShaderNodeBsdfGlossy")
        material_nodes.remove(material_nodes['Principled BSDF'])
        object_material.node_tree.links.new(material_nodes["Material Output"].inputs["Surface"],
                                            bsdf_node.outputs[0])
        # Set material properties
        bsdf_node.inputs["Roughness"].default_value = self._roughness
        bsdf_node.distribution = self._distribution

        material_instance = MaterialInstance(blender_material=object_material,
                                             inputs={"Color": bsdf_node.inputs["Color"]})

        return material_instance

    @property
    def distribution(self):
        return self._distribution


class PrincipledBSDFWireframeMaterial(WireframeMaterial, PrincipledBSDFMaterial):
    def __init__(
            self, wireframe_thickness=0.01, wireframe_color=(0., 0., 0., 1.), **kwargs
    ):
        super().__init__(
            wireframe_thickness=wireframe_thickness, wireframe_color=wireframe_color, **kwargs
        )

    def create_material(self, name: str = "object_material") -> MaterialInstance:
        object_material = bpy.data.materials.new(name=name)
        object_material.use_nodes = True
        object_material.use_backface_culling = self._use_backface_culling
        material_nodes = object_material.node_tree.nodes

        # Create BSDF
        bsdf_node = object_material.node_tree.nodes["Principled BSDF"]

        # Set BSDF properties
        for property_name, blender_name in self._property2blender_mapping.items():
            bsdf_node.inputs[blender_name].default_value = self.__getattribute__("_" + property_name)

        # Create and link wireframe
        self.overlay_wireframe(object_material, bsdf_node)

        # Create internal representation
        material_instance = MaterialInstance(
            blender_material=object_material,
            inputs={
                "Color": bsdf_node.inputs["Base Color"],
                "Alpha": bsdf_node.inputs["Alpha"],
                "Emission": bsdf_node.inputs["Emission"],
                "Emission Strength": bsdf_node.inputs["Emission Strength"]
            }
        )

        return material_instance



from typing import Union

import numpy as np
import trimesh
import torch
import open3d as o3d
from loguru import logger
from scipy.spatial import KDTree
import logging
import time
from pathlib import Path
from skimage.io import imsave

from ..internal.types import Vector3d, Vector4d


def estimate_pc_normals_from_mesh(pc_vertices: np.ndarray, mesh: trimesh.base.Trimesh):
    """Approximate PC per-vertex normals from mesh that is registered to PC.
    For each PC vertex averages vertex normals for 5 nearest mesh vertices.

    Args:
        pc_vertices (np.ndarray): pointcloud vertices (n_vertices x 3)
        mesh (trimesh.base.Trimesh): mesh, registered to the PC;

    Returns:
        np.ndarray: estimated normals n_vertices x 3
    """
    mesh_normals = mesh.vertex_normals
    tree = KDTree(mesh.vertices)
    dd, nn_index = tree.query(pc_vertices, k=5, p=2, workers=-1)
    pc_normals = mesh_normals[nn_index]
    pc_normals = pc_normals.mean(axis=1)
    pc_normals = pc_normals / (np.sqrt(pc_normals ** 2).sum(axis=1, keepdims=True))

    return pc_normals


def estimate_normals_from_pointcloud(pc_vertices: np.ndarray, backend="open3d", device="gpu"):
    """Approximate PC per-vertex normals using algorithms implemented in opend3d or pytorch3d.

    Args:
        pc_vertices (np.ndarray): pointcloud vertices (n_vertices x 3)
        backend (str, optional): backend that is used to estimate normals,
            pytorch3d and open3d are supported (default: "open3d")
        device (str, optional): pytorch device (default: "gpu")

    Returns:
        np.ndarray: estimated normals n_vertices x 3
    """
    if backend == "open3d":
        import open3d as o3d

        pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(pc_vertices))
        pc.estimate_normals(fast_normal_computation=False)
        pc.orient_normals_consistent_tangent_plane(k=10)

        pc_normals = np.asarray(pc.normals).astype(np.float32)
    elif backend == "pytorch3d":
        import torch
        from pytorch3d.ops import estimate_pointcloud_normals

        pc_vertices_th = torch.tensor(pc_vertices, dtype=torch.float, device=device)
        pc_normals_th = estimate_pointcloud_normals(pc_vertices_th.unsqueeze(0), 5).squeeze(0)

        pc_normals = pc_normals_th.cpu().numpy().astype(np.float32)
    else:
        raise NotImplementedError(f"backend {backend} is not supported. Possible backends: open3d, pytorch3d.")

    return pc_normals


def approximate_colors_from_camera(
        camera_viewdir: np.ndarray, vertex_normals: np.ndarray, per_vertex_color: Union[Vector3d, Vector4d],
        back_color: Union[Vector3d, Vector4d] = (0.6, 0.6, 0.6)
):
    """Approximation of visible vertices from camera.
    PC vertices are colored with their initial color only if they are visible from camera (here we use the approximation
    of visibility, by calculating the angle between vertex normal and camera's view direction), otherwise the vertices
    are colored with back_color.

    Args:
        camera_viewdir (np.ndarray): view direction of the camera
        vertex_normals (np.ndarray): per-vertex normals of the point cloud
        per_vertex_color (Union[Vector3d, Vector4d]): colors for the point cloud
        back_color (Union[Vector3d, Vector4d], optional): color for vertices that are not visible from camera. With
            the approximation of visibility, described above (default: (0.6, 0.6, 0.6))

    Returns:
        np.ndarray: new per-vertex coloring with invisible vertices colored in back_color
    """
    # Add alpha
    back_color = np.array(back_color)
    if back_color.shape[0] == 3:
        back_color = np.concatenate((back_color, [1.0]), axis=0)

    # Expand colors if needed
    num_vertices = vertex_normals.shape[0]
    per_vertex_color = np.array(per_vertex_color)
    if per_vertex_color.ndim == 1:
        per_vertex_color = np.repeat(per_vertex_color[np.newaxis], num_vertices, axis=0).astype(np.float32)
    elif per_vertex_color.ndim == 2:
        assert per_vertex_color.shape[0] == num_vertices, \
            "Length of vertex_colors is to be equal to the number of vertices."
    else:
        raise NotImplementedError("Only uniform or per-vertex colors are supported for color approximation.")

    # Add alpha to colors if needed
    if per_vertex_color.shape[1] == 3:
        per_vertex_color = np.concatenate((per_vertex_color, np.ones((num_vertices, 1), np.float32)), axis=1)

    # Add alpha to support transparent back_color
    if per_vertex_color.shape[1] == 3:
        alpha = np.ones((per_vertex_color.shape[0], 1), dtype=np.float32)
        vertex_colors = np.concatenate((per_vertex_color, alpha), axis=1)

    # Compute mask and recolor
    dot_product = (vertex_normals * camera_viewdir[np.newaxis, :]).sum(axis=1)
    back_mask = dot_product > 0.0
    per_vertex_color[back_mask] = back_color[np.newaxis]

    return per_vertex_color


class PCMeshifier:
    torch_dtype = torch.float32
    np_dtype = np.float64

    def __init__(self, subsampled_mesh_faces_count=1_000_000, texture_resolution=(30_000, 30_000), pc_subsample=None,
            max_query_size=2_000, knn=3, gpu_index=None, bpa_radius=None):
        self.subsampled_mesh_faces_count = subsampled_mesh_faces_count
        self.texture_resolution = np.asarray(texture_resolution)
        self.pc_subsample = pc_subsample
        self.max_query_size = max_query_size
        self.knn = knn
        self.gpu_index = gpu_index
        self.bpa_radius = bpa_radius
        self.device = torch.device(f"cuda:{gpu_index}" if gpu_index is not None else "cpu")

    def o3d_mesh_from_pc(self, pc_vertices: np.ndarray, pc_colors: np.ndarray = None, pc_normals: np.ndarray = None):
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(pc_vertices)
        if pc_colors is not None:
            if np.issubdtype(pc_colors.dtype, np.integer):
                pc_colors = pc_colors.astype(np.float64) / 255.
            pc_colors = pc_colors[:, :3].astype(np.float64).copy()
            pcd.colors = o3d.utility.Vector3dVector(pc_colors)
        if pc_normals is None:
            logger.info("pc_normals is None, estimating normals from points")
            pcd.estimate_normals()
            pcd.orient_normals_consistent_tangent_plane(min(len(pc_vertices), 100))
        else:
            pcd.normals = o3d.utility.Vector3dVector(pc_normals)
        logger.info("Generating mesh from PC with BPA (if it takes too long, consider adjusting bpa_radius)")
        stime = time.time()
        if self.bpa_radius is None:
            logger.info("bpa_radius is None, running radius estimation")
            dists = pcd.compute_nearest_neighbor_distance()
            med_dist = np.median(dists)
            bpa_radius = med_dist / np.sqrt(2)
            logger.info(f"Estimated radius: {bpa_radius}")
        else:
            bpa_radius = self.bpa_radius
        bpa_radius_range = o3d.utility.DoubleVector([bpa_radius, bpa_radius * 2])
        bpa_mesh = o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(pcd, bpa_radius_range)
        runtime = time.time() - stime
        logger.info(f"BPA done in {runtime / 3600:.2f} hours ({runtime / 60:.2f} mins)")
        return bpa_mesh

    def o3d_decimate_mesh(self, o3d_mesh: o3d.geometry.TriangleMesh):
        logger.info(f"Performing mesh decimation (to {self.subsampled_mesh_faces_count} faces)")
        stime = time.time()
        dec_mesh = o3d_mesh.simplify_quadric_decimation(self.subsampled_mesh_faces_count)
        dec_mesh.remove_degenerate_triangles()
        dec_mesh.remove_duplicated_triangles()
        dec_mesh.remove_duplicated_vertices()
        dec_mesh.remove_non_manifold_edges()
        runtime = time.time() - stime
        logger.info(f"Decimation done in {runtime / 3600:.2f} hours ({runtime / 60:.2f} mins)")
        return dec_mesh

    def generate_naive_uvmap(self, o3d_mesh: o3d.geometry.TriangleMesh):
        """Will pack faces one after another in the texture, two triangles per square

        Args:
            o3d_mesh (o3d.geometry.TriangleMesh): mesh to generate UV map for

        """
        logger.info("Generating UV map")
        mesh_faces = np.asarray(o3d_mesh.triangles)

        faces_count = len(mesh_faces)
        uv_squares_count = ((faces_count + 1) // 2)
        uv_squares_per_row = np.ceil(np.sqrt(uv_squares_count)).astype(int)
        sq_size = 1. / uv_squares_per_row.astype(np.float64)

        triangle_off = np.array([[(0.1, 0.1), (0.1, 0.8), (0.9, 0.1)],
                                 [(0.9, 0.2), (0.1, 0.9), (0.9, 0.9)]], dtype=self.np_dtype).reshape(1, 6, 2)

        scaled_triangle_off = triangle_off * sq_size
        faces_offsets = np.stack((np.arange(uv_squares_count) // uv_squares_per_row, np.arange(uv_squares_count) % uv_squares_per_row),
                                 axis=1).astype(
            self.np_dtype) * sq_size
        faces_uv = (scaled_triangle_off + faces_offsets[:, None, :]).reshape(uv_squares_count * 2, 3, 2)
        if len(mesh_faces) % 2 == 1:
            faces_uv = faces_uv[:-1]
        logger.info(f"Done, mesh faces count {len(mesh_faces)}, faces uv count {len(faces_uv)}")
        return faces_uv

    def _get_texture_pixels_position_in_3dworld(self, target_texture_coords, mesh_vertices, mesh_faces, uv_map):
        texture_coords_grid = torch.stack([x.reshape(-1) for x in torch.meshgrid(*target_texture_coords, indexing='xy')],
                                          dim=1)  # N,2
        faces_count = len(mesh_faces)
        uv_squares_count = ((faces_count + 1) // 2)
        uv_squares_per_row = np.ceil(np.sqrt(uv_squares_count)).astype(int)

        pix_faces_coords = texture_coords_grid * uv_squares_per_row
        pix_faces_int_coords = torch.floor(pix_faces_coords).long()
        inface_coords = pix_faces_coords - pix_faces_int_coords
        odd_face = (1 - inface_coords[:, 0]) < inface_coords[:, 1]
        pix_faces_inds = (pix_faces_int_coords[:, 0] * uv_squares_per_row + pix_faces_int_coords[:, 1]) * 2 + odd_face
        pix_faces_inds[pix_faces_inds >= faces_count] = faces_count - 1
        # Computing pixels XYZ coords
        pix_faces = mesh_faces[pix_faces_inds]  # N,3
        pix_faces_xyz = mesh_vertices[pix_faces.flatten()].reshape(pix_faces.shape[0], 3, 3)
        pix_faces_uv = uv_map[pix_faces_inds]  # N,3,2
        # Computing barycentric coords
        T = torch.stack([pix_faces_uv[:, 0, :] - pix_faces_uv[:, 2, :], pix_faces_uv[:, 1, :] - pix_faces_uv[:, 2, :]], dim=2)  # N,2,2
        baric_coords2 = torch.linalg.solve(T, texture_coords_grid - pix_faces_uv[:, 2, :])  # N,2
        baric_coords = torch.cat([baric_coords2, 1 - baric_coords2.sum(dim=1)[:, None]], dim=1)  # N,3
        pix_xyz = torch.matmul(baric_coords[:, None, :], pix_faces_xyz)[:, 0, :]  # N,3
        return pix_xyz

    def _generate_texture_block(self, target_texture_coords, mesh_vertices, mesh_faces, uv_map, pc_tree, pc_colors):
        resolution = (len(target_texture_coords[1]), len(target_texture_coords[0]))
        pix_xyz = self._get_texture_pixels_position_in_3dworld(target_texture_coords, mesh_vertices, mesh_faces, uv_map)
        # Querying KDTree
        dists, inds = pc_tree.query(pix_xyz.cpu().numpy(), k=self.knn, workers=-1)
        dists = torch.tensor(dists, dtype=self.torch_dtype, device=self.device)
        inds = torch.tensor(inds, dtype=torch.long, device=self.device)
        # Computing weighed sum
        w = 1. / dists
        w[~torch.isfinite(w)] = 1e6
        w[w > 1e6] = 1e6
        w_sum = w.sum(dim=1)
        target_colors = pc_colors[inds.flatten()].reshape(-1, self.knn, 3)
        pix_colors = (target_colors * w[:, :, None]).sum(dim=1) / w_sum[:, None]
        texture = pix_colors.reshape(resolution + (3,)).clamp(0, 1)
        texture = (texture * 255).cpu().numpy().astype(np.uint8)
        return texture

    def generate_texture_from_pc_block_by_block(self, o3d_mesh: o3d.geometry.TriangleMesh, uv_map: np.ndarray, pc_vertices: np.ndarray,
            pc_colors: np.ndarray,
            query_block_size: int = 2000):

        if self.pc_subsample is not None and self.pc_subsample < len(pc_vertices):
            logger.info(f"Subsampling the cloud to {self.pc_subsample} pts ({self.pc_subsample / len(pc_vertices) * 100:.2f}%)")
            np.random.seed(42)
            sub_inds = np.random.choice(len(pc_vertices), self.pc_subsample, replace=False)
            pc_vertices = pc_vertices[sub_inds]
            pc_colors = pc_colors[sub_inds]

        if np.issubdtype(pc_colors.dtype, np.integer):
            pc_colors = pc_colors.astype(np.float64) / 255.
        pc_colors = pc_colors[:, :3].astype(np.float64).copy()

        logger.info("Creating KDTree for texture generation queries")
        pc_tree = KDTree(pc_vertices)

        mesh_faces = np.asarray(o3d_mesh.triangles)
        mesh_vertices = np.asarray(o3d_mesh.vertices)
        faces_count = len(mesh_faces)
        uv_squares_count = ((faces_count + 1) // 2)

        logger.info("Raising data to device")
        mesh_faces_torch = torch.tensor(mesh_faces, device=self.device)
        mesh_vertices_torch = torch.tensor(mesh_vertices, dtype=self.torch_dtype, device=self.device)
        uv_map_torch = torch.tensor(uv_map, dtype=self.torch_dtype, device=self.device)
        pc_colors_torch = torch.tensor(pc_colors, dtype=self.torch_dtype, device=self.device)

        logger.info("Assigning pixels to faces")
        texture_coords = [(torch.arange(r, dtype=self.torch_dtype, device=self.device) + 0.5) / r for r in self.texture_resolution]
        texture_coords[1] = 1 - texture_coords[1]
        texture_coords_splits = [torch.split(tc, query_block_size) for tc in texture_coords]

        logger.info("Starting texture generation")
        for tex_ind_y, tex_coords_y in enumerate(texture_coords_splits[1]):
            for tex_ind_x, tex_coords_x in enumerate(texture_coords_splits[0]):
                logger.info(f"Generating texture block ({tex_ind_x + 1}, {tex_ind_y + 1}) of "
                            f"({len(texture_coords_splits[0])}, {len(texture_coords_splits[1])})")
                texture_block = self._generate_texture_block([tex_coords_x, tex_coords_y], mesh_vertices_torch,
                                                             mesh_faces_torch, uv_map_torch, pc_tree, pc_colors_torch)
                yield texture_block

    def compute_texture_blocks_count(self, query_block_size):
        blocks_count = np.array(self.texture_resolution + query_block_size - 1) // query_block_size
        return blocks_count

    def generate_texture_from_pc(self, o3d_mesh: o3d.geometry.TriangleMesh, uv_map: np.ndarray,
            pc_vertices: np.ndarray, pc_colors: np.ndarray, query_block_size: int = 2000):
        texture_blocks_count = self.compute_texture_blocks_count(query_block_size)
        curr_row_ind = 0
        all_texture_rows = []
        curr_texture_row = []
        for texture_block in self.generate_texture_from_pc_block_by_block(o3d_mesh, uv_map, pc_vertices, pc_colors, query_block_size):
            curr_texture_row.append(texture_block)
            curr_row_ind += 1
            if curr_row_ind == texture_blocks_count[0]:
                all_texture_rows.append(np.concatenate(curr_texture_row, axis=1))
                curr_row_ind = 0
                curr_texture_row = []
        texture = np.concatenate(all_texture_rows, axis=0)
        return texture


def meshify_pc(pc_vertices, pc_colors, pc_normals=None, subsampled_mesh_faces_count=1_000_000, texture_resolution=(30_000, 30_000), pc_subsample=None,
        max_query_size=2_000, knn=3, gpu_index=None, bpa_radius=None):
    """Turns pointcloud into a subsampled mesh with texture.

    Args:
        pc_vertices (np.ndarray): pointcloud vertices (n_vertices x 3)
        pc_colors (np.ndarray): pointcloud colors (n_vertices x 3)
        pc_normals (np.ndarray, optional): pointcloud normals (n_vertices x 3). If None, normals will be estimated from the pointcloud (default: None)
        subsampled_mesh_faces_count (int, optional): number of faces in the subsampled mesh (default: 1_000_000)
        texture_resolution (tuple, optional): texture resolution (default: (30_000, 30_000))
        pc_subsample (int, optional): number of vertices to subsample from the pointcloud (default: None)
        max_query_size (int, optional): maximum query size for texture generation (default: 2_000)
        knn (int, optional): number of nearest neighbors for texture generation (default: 3)
        gpu_index (int, optional): index of the GPU of texture generation stage. If None, CPU will be used (default: None)
        bpa_radius (float, optional): radius for ball pivoting algorithm for mesh reconstruction.
            If None, radius will be estimated automatically based on PC density

    Returns:
        np.ndarray: vertices of the mesh (V, 3)
        np.ndarray: faces of the mesh (F, 3)
        np.ndarray: per-face UV map of the mesh (F, 3, 2)
        np.ndarray: texture of the mesh (texture_resolution[0], texture_resolution[1], 3)
    """
    meshifier = PCMeshifier(subsampled_mesh_faces_count, texture_resolution, pc_subsample, max_query_size, knn, gpu_index, bpa_radius)
    o3d_mesh = meshifier.o3d_mesh_from_pc(pc_vertices, pc_colors, pc_normals)
    dec_mesh = meshifier.o3d_decimate_mesh(o3d_mesh)
    uv_map = meshifier.generate_naive_uvmap(dec_mesh)
    texture = meshifier.generate_texture_from_pc(dec_mesh, uv_map, pc_vertices, pc_colors)
    mesh_vertices = np.asarray(dec_mesh.vertices)
    mesh_faces = np.asarray(dec_mesh.triangles)
    return mesh_vertices, mesh_faces, uv_map, texture


def meshify_pc_from_file(pc_filepath, output_dir, subsampled_mesh_faces_count=1_000_000, texture_resolution=(30_000, 30_000), pc_subsample=None,
        max_query_size=2_000, knn=3, gpu_index=None, bpa_radius=None):
    """
    Turns pointcloud into a subsampled mesh with texture. Reads pointcloud from a file, writes results to a folder.
        Produces mesh.ply, uv_map.npy and texture.jpg in the target folder

    Args:
        pc_filepath: Path to a pointcloud
        output_dir: Path to a directory which will contain the result
        subsampled_mesh_faces_count (int, optional): number of faces in the subsampled mesh (default: 1_000_000)
        texture_resolution (tuple, optional): texture resolution (default: (30_000, 30_000))
        pc_subsample (int, optional): number of vertices to subsample from the pointcloud (default: None)
        max_query_size (int, optional): maximum query size for texture generation (default: 2_000)
        knn (int, optional): number of nearest neighbors for texture generation (default: 3)
        gpu_index (int, optional): index of the GPU of texture generation stage. If None, CPU will be used (default: None)
        bpa_radius (float, optional): radius in ball pivoting algorithm for mesh reconstruction.
            If None, radius will be estimated automatically based on PC density
    """

    pcd = o3d.io.read_point_cloud(pc_filepath)
    pc_vertices = np.asarray(pcd.points)
    pc_colors = np.asarray(pcd.colors)
    pc_normals = np.asarray(pcd.normals)
    if pc_normals.sum() == 0:
        pc_normals = None

    meshifier = PCMeshifier(subsampled_mesh_faces_count, texture_resolution, pc_subsample, max_query_size,
                            knn, gpu_index, bpa_radius)
    o3d_mesh = meshifier.o3d_mesh_from_pc(pc_vertices, pc_colors, pc_normals)
    dec_mesh = meshifier.o3d_decimate_mesh(o3d_mesh)
    uv_map = meshifier.generate_naive_uvmap(dec_mesh)
    texture = meshifier.generate_texture_from_pc(dec_mesh, uv_map, pc_vertices, pc_colors)

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    o3d.io.write_triangle_mesh(output_dir / "mesh.ply", dec_mesh)
    imsave(output_dir / "texture.jpg", texture, quality=98)
    np.save(output_dir / "uv_map.npy", uv_map)



from typing import Tuple

import bpy

from .base import Material, MaterialInstance


class EmissionMaterial(Material):
    """A class which manages the parameters of GlossyBSDF Blender material.
    Full docs: https://docs.blender.org/manual/en/latest/render/shader_nodes/shader/glossy.html
    """

    def __init__(self, use_backface_culling=True):
        super().__init__(use_backface_culling=use_backface_culling)

    def create_material(self, name: str = "object_material") -> MaterialInstance:
        """Create the Blender material with the parameters stored in the current object

        Args:
            name (str): a unique material name for Blender

        Returns:
            Tuple[bpy.types.Material, bpy.types.ShaderNodeBsdfGlossy]: Blender material and the
                shader node which uses the created material
        """

        object_material = bpy.data.materials.new(name=name)
        object_material.use_nodes = True
        object_material.use_backface_culling = self._use_backface_culling
        material_nodes = object_material.node_tree.nodes

        principled_bsdf = material_nodes['Principled BSDF']
        principled_bsdf.inputs['Base Color'].default_value = (0.0, 0.0, 0.0, 1)
        principled_bsdf.inputs['Alpha'].default_value = 0.5
        principled_bsdf.inputs['Specular'].default_value = 0.0
        principled_bsdf.inputs['Emission Strength'].default_value = 0.5


        emission_node = material_nodes.new("ShaderNodeEmission")
        emission_node.inputs['Strength'].default_value = 10.0

        mix_node = material_nodes.new(type='ShaderNodeMixShader')
        mix_node.inputs['Fac'].default_value = 0

        material_output = material_nodes['Material Output']

        links = object_material.node_tree.links
        links.new(principled_bsdf.outputs['BSDF'], mix_node.inputs[1])
        links.new(emission_node.outputs['Emission'], mix_node.inputs[2])
        links.new(mix_node.outputs['Shader'], material_output.inputs['Surface'])

        material_instance = MaterialInstance(blender_material=object_material,
                                      #       inputs={"Color": emission_node.inputs["Color"]})
                                             inputs={"Color": principled_bsdf.inputs["Emission"]})
        return material_instance

    @property
    def distribution(self):
        return self._distribution



"""
MIT License

Copyright (c) 2018 Sebastian Bullinger
Copyright (c) 2021 Vladimir Guzov and Ilia Petrov

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import warnings
from dataclasses import dataclass
from typing import List, Optional, Dict, Tuple

import bpy
import numpy as np
from mathutils import Vector

from .base import Renderable
from ..colors import VertexColors, UniformColors
from ..colors.base import ColorsMetadata, Colors
from ..internal.texture import compute_particle_color_texture
from ..materials.base import Material, MaterialInstance


@ dataclass
class ParticleMetadata:
    """Helper class that stores pointers to blender objects (colors, bsdf and material nodes),
    connected to each Particle System object
    """
    material_instance: Optional[MaterialInstance]  # Material instance for the particle system
    extra_color_nodes: Tuple[bpy.types.ShaderNode]  # Extra nodes in the material node tree necessary for proper color computation
    vertex_offset: int  # index of a starting vertex
    num_particles: int  # number of particles in particle collection
    texture_image: Optional[bpy.types.Image]  # texture image for per-vertex colors


class PointCloud(Renderable):
    """Basic point cloud consisting of vertices, supports uniform (UniformColors) and per-vertex (VertexColors) coloring.
    Uses Blender-Photogrammetry-Importer to handle point clouds as Blender particle systems objects

    Properties:
        emit_shadow (bool, optional): control whether particles representing the point cloud will emit shadow from
            any light source in the scene. This property may be turned off if the particle_emission_strength is big
            enough to avoid artifacts.
    """
    def __init__(
            self,
            vertices: np.ndarray,
            tag: str,
            point_size: float = 0.006,
            base_primitive: str = "CUBE",
            particle_emission_strength: int = 1,
            **kwargs
    ):
        """Creates Blender Collection that represent given point cloud. Code for creation particle systems for
        representing the point clouds is borrowed from https://github.com/SBCV/Blender-Addon-Photogrammetry-Importer

        Args:
            vertices (np.ndarray): point cloud vertices
            material (Material): Material instance
            colors (Colors): VertexColors or UniformColors instance
            point_size (float, optional): size of a primitive, representing each vertex (default: 0.006)
            base_primitive (str, optional): type of primitive for representing each point
                (possible values are PLANE, CUBE, SPHERE, default: CUBE)
            particle_emission_strength (int, optional): strength of the emission from each primitive. This is used to
                increase realism. Values <= 0 turn emission off, values > 0 set the power of emission (default: 1)
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created collection in Blender
        """

        # internal variables and constants
        # This is a constraint imposed by open bug in blender https://developer.blender.org/T81103
        self._max_particles: int = 10000
        self._particle_collections = dict()
        self._point_size: float = point_size
        self._base_primitive: str = base_primitive.upper()
        self.particle_emission_strength: int = particle_emission_strength
        self._point_cloud_object_names: List[str] = list()
        self._particle_material_names: List[str] = list()

        self.num_vertices: int = 0
        self._particle_metadata: Dict[str, ParticleMetadata] = dict()  # particle_object_name: ParticleMetadata
        self._colors_metadata: Optional[ColorsMetadata] = None

        collection = self._blender_create_collection(vertices, tag)
        super().__init__(**kwargs, blender_object=collection, tag=tag)

    def update_vertices(
            self,
            vertices: np.ndarray
    ):
        """Updates pc vertices coordinates

        Args:
            vertices (np.ndarray): new coordinates for point cloud vertices
        """
        assert self.num_vertices == len(vertices), \
            f"Number of vertices should be the same (expected {self.num_vertices}, got {len(vertices)})"
        for subset_ind, offset in enumerate(range(0, self.num_vertices, self._max_particles)):
            points_subset = vertices[offset: offset + self._max_particles]

            # update PC object
            pc_object = self._blender_object.all_objects[f"Particle_{subset_ind}_PC"]
            for vert_ind, vert in enumerate(pc_object.vertices):
                vert.co = points_subset[vert_ind]
            pc_object.update()

    # Getter and setter for emit_shadow: this property may be turned off if the particle_emission_strength
    # is big enough to avoid artifacts.
    @property
    def emit_shadow(self):
        val = None
        for particle_obj_name in self._particle_metadata.keys():
            particle_obj = self._blender_object.all_objects[particle_obj_name]
            curr_val = particle_obj.cycles_visibility.shadow
            if val is None:
                val = curr_val
            elif val != curr_val:
                return None
        return val

    @emit_shadow.setter
    def emit_shadow(self, val: bool):
        for particle_obj_name in self._particle_metadata.keys():
            particle_obj = self._blender_object.all_objects[particle_obj_name]
            particle_obj.cycles_visibility.shadow = val

    # ===================================================== OBJECT =====================================================
    def _blender_create_collection(
            self,
            vertices: np.ndarray,
            tag: str
    ) -> bpy.types.Collection:
        """Creates Blender collection of particle systems, that represent point cloud

        Args:
            vertices (np.ndarray): all points in the point cloud
            tag (str): a name for a blender collection

        Returns:
            bpy.types.Collection: a collection of particle systems representing the point cloud
        """
        new_collection = bpy.data.collections.new(f"Particle System {tag}")
        bpy.context.collection.children.link(new_collection)

        self.point_cloud_obj_list = []
        self.num_vertices = len(vertices)
        for index, vertex_start in enumerate(range(0, self.num_vertices, self._max_particles)):
            # This name is used is a unique identifier for
            # the internal dictionary with metadata self._particle_metadata
            particle_obj_name = f"Particle_{index}"
            # particle_material_name = f"Particle_{index}_Material"
            point_cloud_obj_name = f"Particle_{index}_PC"

            self._point_cloud_object_names.append(point_cloud_obj_name)
            points_subset = vertices[vertex_start: vertex_start + self._max_particles]
            self._particle_metadata[particle_obj_name] = ParticleMetadata(
                material_instance=None,
                extra_color_nodes=tuple(),
                vertex_offset=vertex_start,
                num_particles=len(points_subset),
                texture_image=None
            )

            particle_obj = self._add_particle_obj(
                particle_obj_name=particle_obj_name,
                mesh_type=self._base_primitive,
                point_size=self._point_size,
                blender_collection=new_collection
            )
            point_cloud_obj = self._add_particle_system_obj(
                coords=points_subset,
                particle_obj=particle_obj,
                point_cloud_obj_name=point_cloud_obj_name,
                blender_collection=new_collection,
            )
            self.point_cloud_obj_list.append(point_cloud_obj)

        return new_collection

    def _blender_remove_object(self):
        """Removes the object from Blender scene
        """
        self._blender_clear_colors()
        self._blender_clear_material()
        super()._blender_remove_object()

    @staticmethod
    def _add_particle_obj(
            particle_obj_name: str,
            mesh_type: str,
            point_size: float,
            blender_collection: bpy.types.Collection
    ) -> bpy.types.Object:
        """Creates particle that will be used in particle system to represent vertices in the point cloud

        Args:
            particle_obj_name (str): name of a Blender primitive object to be created
            mesh_type (str): type of primitive for representing each point (possible values are PLANE, CUBE, SPHERE)
            point_size (float): size of a primitive
            blender_collection (bpy.types.Collection): a Blender collection for storing the primitive

        Returns:
            bpy.types.Object: a Blender primitive that is used to represent point cloud vertex
        """
        # The default size of elements added with
        #   primitive_cube_add, primitive_uv_sphere_add, etc. is (2,2,2)
        point_scale = point_size * 0.5
        bpy.ops.object.select_all(action="DESELECT")
        if mesh_type == "PLANE":
            bpy.ops.mesh.primitive_plane_add(size=point_scale)
        elif mesh_type == "CUBE":
            bpy.ops.mesh.primitive_cube_add(size=point_scale)
        elif mesh_type == "SPHERE":
            bpy.ops.mesh.primitive_uv_sphere_add(radius=point_scale)
        else:
            bpy.ops.mesh.primitive_uv_sphere_add(radius=point_scale)
        particle_obj = bpy.context.object
        bpy.context.view_layer.objects.active = particle_obj
        bpy.ops.object.mode_set(mode='OBJECT')
        # Purposely repeating object name setting 2 times - for some primitives (cubes) setting it just ones sometimes doesn't work
        particle_obj.name = particle_obj_name
        particle_obj.name = particle_obj_name
        blender_collection.objects.link(particle_obj)
        bpy.context.collection.objects.unlink(particle_obj)

        return particle_obj

    @staticmethod
    def _add_particle_system_obj(
            coords: np.ndarray,
            particle_obj: bpy.types.Object,
            point_cloud_obj_name: str,
            blender_collection: bpy.types.Collection
    ) -> bpy.types.Object:
        """Creates particle system that represents a part of the point cloud (up to 10k vertices)
        using a given primitive

        Args:
            coords (np.ndarray): coordinates of point cloud vertices
            particle_obj (bpy.types.Object): a Blender primitive object that represents point cloud vertex
            point_cloud_obj_name (str): unique identifier of Blender particle system object tho be created
            blender_collection (bpy.types.Collection): a Blender collection for storing the particle system

        Returns:
            bpy.types.Object: Blender particle system object
        """
        point_cloud_mesh = bpy.data.meshes.new(point_cloud_obj_name)
        point_cloud_mesh.update()
        point_cloud_mesh.validate()
        point_cloud_mesh.from_pydata(coords, [], [])

        point_cloud_obj = bpy.data.objects.new(point_cloud_obj_name, point_cloud_mesh)
        blender_collection.objects.link(point_cloud_obj)
        point_cloud_obj.select_set(state=True)

        if bpy.context.view_layer.objects.active is None or bpy.context.view_layer.objects.active.mode == "OBJECT":
            bpy.context.view_layer.objects.active = point_cloud_obj

        if len(point_cloud_obj.particle_systems) == 0:
            point_cloud_obj.modifiers.new("particle sys", type="PARTICLE_SYSTEM")
            particle_sys = point_cloud_obj.particle_systems[0]
            settings = particle_sys.settings
            settings.type = "HAIR"
            settings.use_advanced_hair = True
            settings.emit_from = "VERT"
            settings.count = len(coords)
            # The final object extent is hair_length * obj.scale
            settings.hair_length = 100  # This must not be 0
            settings.use_emit_random = False
            settings.render_type = "OBJECT"
            settings.instance_object = particle_obj
            settings.use_rotations = True
            settings.rotation_mode = "GLOB_X"
        return point_cloud_obj
    # ================================================== END OF OBJECT =================================================

    # ==================================================== MATERIAL ====================================================
    def update_material(
            self,
            material: Material
    ):
        """Updates object material properties, sets Blender structures accordingly

        Args:
            material (Material): target material
        """
        if not isinstance(material, Material):
            assert len(material) == 1, "Only one material can be provided for the point cloud"
            material = material[0]
        self._blender_clear_material()
        self._blender_set_material(material)

    def _blender_set_material(
            self,
            material: Material
    ):
        """Constructs material node, recreates color node if needed

        Args:
            material (Material): target material
        """
        for particle_obj_name, metadata in self._particle_metadata.items():
            material_instance = material.create_material(name=f"{particle_obj_name}_Material")
            blender_material = material_instance.blender_material
            metadata.material_instance = material_instance
            particle_obj = self._blender_object.all_objects[particle_obj_name]
            particle_obj.data.materials.append(blender_material)

        self._blender_create_colors_node()
        self._blender_link_color2material()

    def _blender_clear_material(
            self,
    ):
        """Clears Blender material node and nodes connected to it
        """
        for particle_obj_name, metadata in self._particle_metadata.items():
            if metadata.material_instance is not None:
                material_instance = metadata.material_instance
                material_nodes = material_instance.blender_material.node_tree.nodes
                blender_material = material_instance.blender_material
                for material_node in material_nodes:
                    blender_material.node_tree.nodes.remove(material_node)
                particle_obj = self._blender_object.all_objects[particle_obj_name]
                particle_obj.data.materials.clear()
                blender_material.user_clear()
                bpy.data.materials.remove(blender_material)

                metadata.material_instance = None
                metadata.extra_color_nodes = tuple()

    # ================================================ END OF MATERIAL =================================================

    # ===================================================== COLORS =====================================================
    def update_colors(
            self,
            colors: Colors
    ):
        """Updates object color properties, sets Blender structures accordingly

        Args:
            colors (Colors): target colors information
        """
        if not isinstance(colors, Colors):
            assert len(colors) == 1, "Only one color can be provided for the point cloud"
            colors = colors[0]
        self._blender_clear_colors()
        self._blender_set_colors(colors)

    def _blender_set_colors(
            self,
            colors: Colors
    ):
        """Remembers current color properties, builds a color node for material

        Args:
            colors (Colors): target colors information
        """
        self._colors_metadata = colors.metadata

        # Create artificial textures if we have VertexColors
        if self._colors_metadata.type == VertexColors:
            for particle_obj_name, metadata in self._particle_metadata.items():
                vertex_offset = metadata.vertex_offset
                vertex_colors_subset = colors.vertex_colors[vertex_offset: vertex_offset + self._max_particles]
                metadata.texture_image = compute_particle_color_texture(vertex_colors_subset)

        self._blender_create_colors_node()
        self._blender_link_color2material()

    def _blender_clear_colors(
            self,
    ):
        """Clears Blender color node and erases node constructor
        """
        for particle_obj_name, metadata in self._particle_metadata.items():
            material_instance = metadata.material_instance
            if material_instance.colors_node is not None:
                blender_material = material_instance.blender_material
                blender_material.node_tree.nodes.remove(material_instance.colors_node)
                for extra_node in metadata.extra_color_nodes:
                    blender_material.node_tree.nodes.remove(extra_node)

                # Clear artificially created texture 
                if metadata.texture_image is not None:
                    # Remove only if the image has no users.
                    if not metadata.texture_image.users:
                        bpy.data.images.remove(metadata.texture_image)
                material_instance.colors_node = None
                metadata.extra_color_nodes = tuple()
                metadata.texture_image = None
                self._colors_metadata = None

    def _blender_create_colors_node(self):
        """Creates color node using previously set builder
        """
        if self._colors_metadata is not None:
            for particle_obj_name, metadata in self._particle_metadata.items():
                material_instance = metadata.material_instance
                blender_material = material_instance.blender_material

                if self._colors_metadata.type == UniformColors:
                    colors_node = blender_material.node_tree.nodes.new('ShaderNodeRGB')
                    colors_node.outputs[0].default_value = Vector(self._colors_metadata.color.tolist() + [1.]).to_4d()
                    material_instance.colors_node = colors_node
                elif self._colors_metadata.type == VertexColors:
                    particle_color_node = blender_material.node_tree.nodes.new("ShaderNodeTexImage")
                    particle_color_node.interpolation = "Closest"
                    particle_color_node.image = metadata.texture_image
                    particle_info_node = blender_material.node_tree.nodes.new("ShaderNodeParticleInfo")

                    # Idea: we use the particle idx to compute a texture coordinate
                    # Shift the un-normalized texture coordinate by a half pixel
                    shift_half_pixel_node = blender_material.node_tree.nodes.new("ShaderNodeMath")
                    shift_half_pixel_node.operation = "ADD"
                    blender_material.node_tree.links.new(
                        particle_info_node.outputs["Index"],
                        shift_half_pixel_node.inputs[0],
                    )
                    shift_half_pixel_node.inputs[1].default_value = 0.5

                    # Compute normalized texture coordinates (value between 0 and 1)
                    # by dividing by the number of particles
                    divide_node = blender_material.node_tree.nodes.new("ShaderNodeMath")
                    divide_node.operation = "DIVIDE"
                    blender_material.node_tree.links.new(
                        shift_half_pixel_node.outputs["Value"],
                        divide_node.inputs[0],
                    )
                    divide_node.inputs[1].default_value = metadata.num_particles

                    # Compute texture coordinate (x axis corresponds to particle idx)
                    shader_node_combine = blender_material.node_tree.nodes.new("ShaderNodeCombineXYZ")
                    blender_material.node_tree.links.new(
                        divide_node.outputs["Value"], shader_node_combine.inputs["X"]
                    )
                    blender_material.node_tree.links.new(
                        shader_node_combine.outputs["Vector"],
                        particle_color_node.inputs["Vector"],
                    )
                    material_instance.colors_node = particle_color_node
                    metadata.extra_color_nodes = (particle_info_node, shift_half_pixel_node, divide_node, shader_node_combine)
                else:
                    raise NotImplementedError(f"Unsupported colors class '{self._colors_metadata.type}'")

    def _blender_link_color2material(
            self
    ):
        """Links color and material nodes, additionally adds emission to particle color if needed
        """
        for particle_obj_name, metadata in self._particle_metadata.items():
            material_instance = metadata.material_instance
            if material_instance is not None:
                colors_node = material_instance.colors_node

                if colors_node is not None:
                    blender_material = material_instance.blender_material
                    colors_metadata = self._colors_metadata
                    blender_material.node_tree.links.new(material_instance.inputs['Color'],
                                                         material_instance.colors_node.outputs['Color'])
                    material_instance.inputs['Color'].default_value = [1.0, 0.0, 0.0, 1.0]

                    if colors_metadata.has_alpha:
                        if 'Alpha' in material_instance.inputs:
                            if colors_metadata.type is UniformColors:
                                material_instance.inputs['Alpha'].default_value = colors_metadata.color[3]
                            else:
                                blender_material.node_tree.links.new(material_instance.inputs['Alpha'],
                                                                     material_instance.colors_node.outputs['Alpha'])
                        else:
                            warnings.warn("This material does not support transparency; alpha color channel is ignored")

                    if self.particle_emission_strength > 0:
                        if 'Emission' in material_instance.inputs:
                            blender_material.node_tree.links.new(
                                colors_node.outputs["Color"],
                                material_instance.inputs["Emission"],
                            )
                            material_instance.inputs["Emission Strength"].default_value = self.particle_emission_strength
                        else:
                            warnings.warn("This material does not support light emission; emission is ignored")

    # ================================================== END OF COLORS =================================================



class Singleton(type):
    """An implementation of the singleton metaclass.
    This is a metaclass which allows only one instance of the class to exist.
    Useful to handle an object which should not be duplicated (such as Blender scene).
    """
    _instances = dict()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
            return cls._instances[cls]
        else:
            raise RuntimeError(f"Only one instance of class {cls.__name__} is allowed")



import functools
import os.path as osp

import cv2
import numba
import numpy as np
from blendipose.blendify import scene
from blendipose.blendify.colors import TextureColors, VertexUV


def add_billboard(
    texture, material, scale, rotation=None, rotation_mode='quaternionWXYZ', translation=(0, 0, 0)
):
    uv_map = VertexUV(np.array([[0, 0], [0, 1], [1, 1], [1, 0]]))
    colors = TextureColors(texture, uv_map, interpolation='Closest')
    verts = np.array([[-1, -1, 0], [-1, 1, 0], [1, 1, 0], [1, -1, 0]]) * scale / 2
    faces = np.array([[0, 1, 2], [0, 2, 3]])
    return scene.renderables.add_mesh(
        verts,
        faces,
        material,
        colors,
        rotation=rotation,
        rotation_mode=rotation_mode,
        translation=translation,
    )


#
# def make_emissive(obj):
#     mat = obj.data.materials[0]  # Get the first material
#     nodes = mat.node_tree.nodes
#     links = mat.node_tree.links
#
#     # Clear existing nodes (if starting fresh)
#     # nodes.clear()
#
#     # Create Emission Shader
#     emission = nodes.new(type='ShaderNodeEmission')
#     emission.location = -200, 300
#
#     # Create Mix Shader
#     mix_shader = nodes.new(type='ShaderNodeMixShader')
#     mix_shader.location = 50, 300
#
#     # Get Principled BSDF and Material Output
#     principled_bsdf = next(node for node in nodes if node.type == 'BSDF_PRINCIPLED')
#     material_output = next(node for node in nodes if node.type == 'OUTPUT_MATERIAL')
#
#     texture_node = next((node for node in nodes if node.type == 'TEX_IMAGE' and any(
#         link.to_node == principled_bsdf for link in node.outputs['Color'].links)), None)
#
#     if texture_node:
#         # Disconnect the texture node from the base color
#         for link in texture_node.outputs['Color'].links:
#             if link.to_node == principled_bsdf and link.to_socket.name == 'Base Color':
#                 links.remove(link)
#                 pass
#
#     principled_bsdf.inputs['Base Color'].default_value = (0.0, 0.0, 0.0, 1)
#     emission.inputs['Strength'].default_value = 10.0
#     # Connect Nodes
#     links.new(texture_node.outputs['Color'], emission.inputs['Color'])
#     links.new(emission.outputs['Emission'], mix_shader.inputs[2])
#
#     links.new(principled_bsdf.outputs['BSDF'], mix_shader.inputs[1])
#     links.new(mix_shader.outputs['Shader'], material_output.inputs['Surface'])
#
#     mix_shader.inputs['Fac'].default_value = 0.05


# def alpha_blend(im, overlay_rgba):
#     alpha = overlay_rgba[:, :, 3:].astype(np.float32) / 255
#     return (np.asarray(im, np.float32) * (1 - alpha) + overlay_rgba[:, :, :3].astype(
#         np.float32) * alpha).astype(np.uint8)
#
#
# def blend_with_background(img: np.ndarray, bkg_color=(1., 1., 1.)) -> np.ndarray:
#     """Blend the RGBA image with uniform colored background, return RGB image
#
#     Args:
#         img: RGBA foreground image
#         bkg_color: RGB uniform background color (default is white)
#
#     Returns:
#         np.ndarray: RGB image blended with background
#     """
#     bkg_color = np.array(bkg_color)
#     if img.dtype == np.uint8:
#         bkg_color_uint8 = (bkg_color * 255).astype(np.uint8)
#         alpha = img[:, :, 3:4].astype(np.int32)
#         img_with_bkg = ((img[:, :, :3] * alpha + bkg_color_uint8[None, None, :] * (
#                     255 - alpha)) // 255).astype(np.uint8)
#     else:
#         alpha = img[:, :, 3:4]
#         img_with_bkg = (img[:, :, :3] * alpha + bkg_color[None, None, :] * (1. - alpha))
#     return img_with_bkg


def resize_image(im, dst_shape):
    if im.shape[:2] == dst_shape[:2]:
        return im
    interp = cv2.INTER_LINEAR if dst_shape[0] > im.shape[0] else cv2.INTER_AREA
    return cv2.resize(im, (dst_shape[1], dst_shape[0]), interpolation=interp)


def smootherstep(x, x0=0, x1=1, y0=0, y1=1):
    a = np.clip((x - x0) / (x1 - x0), 0.0, 1.0)
    b = a**3 * (a * (a * 6.0 - 15.0) + 10.0)
    return y0 + b * (y1 - y0)


def replace_extension(path, new_ext):
    return osp.splitext(path)[0] + new_ext


@functools.lru_cache
@numba.njit(error_model='numpy', cache=True)
def get_srgb_decoder_lut(encoded_dtype=np.uint8):
    maxval = np.iinfo(encoded_dtype).max
    length = int(maxval) + 1
    lut = np.zeros(length, np.float64)
    for i in numba.prange(length):
        x = i / maxval
        if x <= 0.04045:
            lut[i] = x / 12.92
        else:
            lut[i] = ((x + 0.055) / 1.055) ** 2.4
        if lut[i] < 0:
            lut[i] = 0
        elif lut[i] > 1:
            lut[i] = 1
    return (lut * (1 << 16 - 1)).astype(np.uint16)


@functools.lru_cache
@numba.njit(error_model='numpy', cache=True)
def get_srgb_encoder_lut(encoded_dtype=np.uint8):
    lut = np.zeros(1 << 16, np.float64)
    for i in numba.prange(1 << 16):
        x = i / (1 << 16 - 1)
        if x <= 0.0031308:
            lut[i] = x * 12.92
        else:
            lut[i] = 1.055 * x ** (1 / 2.4) - 0.055
        if lut[i] < 0:
            lut[i] = 0
        elif lut[i] > 1:
            lut[i] = 1

    maxval = np.iinfo(encoded_dtype).max
    return (lut * maxval).astype(encoded_dtype)


def alpha_blend(overlay_rgba, background=255, use_srgb_lut=True, dst=None, dtype=np.uint8):
    if np.isscalar(background):
        background = np.array([background, background, background], dtype)
    else:
        background = np.asarray(background, dtype)

    if background.ndim == 1:
        if use_srgb_lut:
            return _alpha_blend_singlecolor(
                overlay_rgba,
                background,
                get_srgb_encoder_lut(dtype),
                get_srgb_decoder_lut(dtype),
                dst,
                dtype,
            )
        else:
            return _alpha_blend_singlecolor_nolut(overlay_rgba, background, dst, dtype)
    else:
        if use_srgb_lut:
            return _alpha_blend(
                overlay_rgba,
                background,
                get_srgb_encoder_lut(dtype),
                get_srgb_decoder_lut(dtype),
                dst,
                dtype,
            )
        else:
            return _alpha_blend_nolut(overlay_rgba, background, dst, dtype)


@numba.njit(error_model='numpy', cache=True)
def _alpha_blend(overlay_rgba, im, srgb_encoder_lut, srgb_decoder_lut, dst=None, dtype=np.uint8):
    if dst is None:
        dst = np.empty((overlay_rgba.shape[0], overlay_rgba.shape[1], 3), dtype)

    rgba_flat = overlay_rgba.reshape(-1, 4)
    rgb_flat = im.reshape(-1, 3)
    dst_flat = dst.reshape(-1, 3)

    maxval = np.iinfo(dtype).max

    for i in numba.prange(rgba_flat.shape[0]):
        alpha = np.float32(rgba_flat[i, 3]) / maxval
        for c in range(3):
            rgb_linear1 = srgb_decoder_lut[rgba_flat[i, c]]
            rgb_linear2 = srgb_decoder_lut[rgb_flat[i, c]]
            combined_linear = (
                np.float32(rgb_linear2)
                + (np.float32(rgb_linear1) - np.float32(rgb_linear2)) * alpha
            )
            if combined_linear < 0:
                combined_linear = 0
            elif combined_linear > 65535:
                combined_linear = 65535

            dst_flat[i, c] = srgb_encoder_lut[np.uint16(combined_linear)]
    return dst


@numba.njit(error_model='numpy', cache=True)
def _alpha_blend_singlecolor(
    overlay_rgba, color, srgb_encoder_lut, srgb_decoder_lut, dst=None, dtype=np.uint8
):
    if dst is None:
        dst = np.empty((overlay_rgba.shape[0], overlay_rgba.shape[1], 3), dtype)

    rgba_flat = overlay_rgba.reshape(-1, 4)
    rgb_linear2_float = srgb_decoder_lut[color].astype(np.float32)
    dst_flat = dst.reshape(-1, 3)

    maxval = np.iinfo(dtype).max

    for i in numba.prange(rgba_flat.shape[0]):
        alpha = np.float32(rgba_flat[i, 3]) / maxval
        for c in range(3):
            rgb_linear1 = srgb_decoder_lut[rgba_flat[i, c]]
            combined_linear = (
                rgb_linear2_float[c] + (np.float32(rgb_linear1) - rgb_linear2_float[c]) * alpha
            )
            if combined_linear < 0:
                combined_linear = 0
            elif combined_linear > 65535:
                combined_linear = 65535

            dst_flat[i, c] = srgb_encoder_lut[np.uint16(combined_linear)]
    return dst


@numba.njit(error_model='numpy', cache=True)
def _alpha_blend_singlecolor_nolut(overlay_rgba, color, dst=None, dtype=np.uint8):
    if dst is None:
        dst = np.empty((overlay_rgba.shape[0], overlay_rgba.shape[1], 3), dtype)

    rgba_flat = overlay_rgba.reshape(-1, 4)
    rgb2_float = color.astype(np.float32)
    dst_flat = dst.reshape(-1, 3)

    maxval = np.iinfo(dtype).max

    for i in numba.prange(rgba_flat.shape[0]):
        alpha = np.float32(rgba_flat[i, 3]) / maxval
        for c in range(3):
            rgb1 = rgba_flat[i, c]
            combined = rgb2_float[c] + (np.float32(rgb1) - rgb2_float[c]) * alpha
            if combined < 0:
                combined = 0
            elif combined > maxval:
                combined = maxval
            dst_flat[i, c] = dtype(combined)
    return dst


@numba.njit(error_model='numpy', cache=True)
def _alpha_blend_nolut(overlay_rgba, im, dst=None, dtype=np.uint8):
    if dst is None:
        dst = np.empty((overlay_rgba.shape[0], overlay_rgba.shape[1], 3), dtype)

    rgba_flat = overlay_rgba.reshape(-1, 4)
    rgb_flat = im.reshape(-1, 3)
    dst_flat = dst.reshape(-1, 3)

    maxval = np.iinfo(dtype).max

    for i in numba.prange(rgba_flat.shape[0]):
        alpha = np.float32(rgba_flat[i, 3]) / maxval
        for c in range(3):
            rgb1 = rgba_flat[i, c]
            rgb2 = rgb_flat[i, c]
            combined = np.float32(rgb2) + (np.float32(rgb1) - np.float32(rgb2)) * alpha
            if combined < 0:
                combined = 0
            elif combined > maxval:
                combined = maxval

            dst_flat[i, c] = dtype(np.rint(combined))
    return dst


MM_TO_UNIT = 1 / 1000
UNIT_TO_MM = 1000
WORLD_UP = np.array([0, 0, 1])
WORLD_TO_BLENDERWORLD_ROTATION_MAT = np.array([[1, 0, 0], [0, 0, 1], [0, -1, 0]], dtype=np.float32)

# Our convention follows OpenCV, where x is right, y is down, and z is forward
# Blender's convention is x is right, y is up, and z is backward
CAM_TO_BLENDERCAM_ROTATION_MAT = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]], dtype=np.float32)


def rotation_mat(up):
    up = unit_vector(up)
    rightlike = np.array([1, 0, 0])
    if np.allclose(up, rightlike):
        rightlike = np.array([0, 1, 0])

    forward = unit_vector(np.cross(up, rightlike))
    right = np.cross(forward, up)

    # In Blender, the world coordinate system is right-handed, with z pointing up
    return np.row_stack([right, forward, up])


def unit_vector(vectors, axis=-1):
    norm = np.linalg.norm(vectors, axis=axis, keepdims=True)
    return vectors / norm


def world_to_blender(points):
    points = np.asarray(points)
    if points.ndim == 1:
        return WORLD_TO_BLENDERWORLD_ROTATION_MAT @ points * MM_TO_UNIT
    else:
        return points @ WORLD_TO_BLENDERWORLD_ROTATION_MAT.T * MM_TO_UNIT


def blender_to_world(points):
    points = np.asarray(points)
    if points.ndim == 1:
        return WORLD_TO_BLENDERWORLD_ROTATION_MAT.T @ points * UNIT_TO_MM
    else:
        return points @ WORLD_TO_BLENDERWORLD_ROTATION_MAT * UNIT_TO_MM


def set_world_up(world_up):
    global WORLD_UP
    WORLD_UP = np.asarray(world_up)
    global WORLD_TO_BLENDERWORLD_ROTATION_MAT
    WORLD_TO_BLENDERWORLD_ROTATION_MAT = rotation_mat(WORLD_UP)



from abc import abstractmethod

import bpy
import numpy as np
from scipy.spatial import transform

from ..internal.positionable import Positionable
from ..internal.types import Vector2di


class Camera(Positionable):
    """Base class for PerspectiveCamera and OrthographicCamera, implementing shared functionality
    """
    @abstractmethod
    def __init__(
        self,
        resolution: Vector2di,
        near: float = 0.1,
        far: float = 100,
        tag: str = 'camera',
        **kwargs
    ):
        """Implements the creation of camera in Blender. Called by child classes

        Args:
            resolution (Vector2di): (w, h), the resolution of the resulting image
            near (float, optional): Camera near clipping distance (default: 0.1)
            far (float, optional): Camera far clipping distance (default: 100)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        camera_object = self._blender_create_camera(tag)
        super().__init__(**kwargs, tag=tag, blender_object=camera_object)
        camera_object.data.sensor_fit = 'HORIZONTAL'
        camera_object.data.sensor_width = resolution[0]
        camera_object.data.sensor_height = resolution[1]
        self.near = near
        self.far = far

        self._resolution = np.array(resolution)

    def _blender_create_camera(self, tag):
        bpy.ops.object.camera_add()
        camera_object = bpy.data.objects['Camera']
        camera_object.name = tag
        return camera_object

    @property
    def resolution(self) -> np.ndarray:
        return self._resolution

    @property
    def blender_camera(self) -> bpy.types.Object:
        return self._blender_object

    @abstractmethod
    def distance2depth(self, distmap: np.ndarray) -> np.ndarray:
        """Convert map of camera ray lengths (distmap) to map of distances to image plane (depthmap)

        Args:
            distmap (np.ndarray): Distance map

        Returns:
            np.ndarray: Depth map
        """
        pass

    @property
    def near(self) -> float:
        return self.blender_camera.data.clip_start

    @near.setter
    def near(self, val: float):
        self.blender_camera.data.clip_start = val

    @property
    def far(self) -> float:
        return self.blender_camera.data.clip_end

    @far.setter
    def far(self, val: float):
        self.blender_camera.data.clip_end = val

    def get_camera_viewdir(self):
        # Default blender camera: up is aligned with +y, ray: (0,0,-1)
        camera_ray = np.array([0, 0, -1], dtype=np.float32)

        # scipy quat is [x, y, z, w], while ours is [w, x, y, z]
        rotation = transform.Rotation.from_quat(np.roll(self.quaternion, -1))
        camera_ray = rotation.apply(camera_ray)
        return camera_ray

    def _update_position(self):
        super()._update_position()



from .bsdf import PrincipledBSDFMaterial, PrincipledBSDFWireframeMaterial


class MetalMaterial(PrincipledBSDFMaterial):
    def __init__(
            self, roughness=0.55, metallic=0.9, clearcoat=0.0, specular=0.5,
            clearcoat_roughness=0.0, **kwargs
    ):
        super().__init__(
            roughness=roughness, metallic=metallic, clearcoat=clearcoat, specular=specular,
            clearcoat_roughness=clearcoat_roughness, **kwargs
        )


class MetalWireframeMaterial(PrincipledBSDFWireframeMaterial):
    def __init__(
            self, roughness=0.55, metallic=0.9,
            clearcoat=0.0, specular=0.5, clearcoat_roughness=0.0,
            wireframe_thickness=0.01, wireframe_color=(0., 0., 0., 1.), **kwargs
    ):
        super().__init__(
            wireframe_thickness=wireframe_thickness, wireframe_color=wireframe_color,
            roughness=roughness, metallic=metallic, clearcoat=clearcoat, specular=specular,
            clearcoat_roughness=clearcoat_roughness, **kwargs
        )



from abc import ABC, abstractmethod

import bpy
import numpy as np
from scipy.spatial.transform import Rotation

from ..internal.types import BlenderGroup, Vector3d, Vector4d, RotationParams


class Positionable(ABC):
    """Base class for all classes that wrap Blender objects with location in space (Camera, Light, Renderable)
    """

    @abstractmethod
    def __init__(
            self,
            tag: str,
            blender_object: BlenderGroup,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0)
    ):
        """Sets initial position of the Blender object and stores it. Called from child classes

        Args:
            tag (str): name of the object in Blender that was created in child class
            blender_object (bpy.types.Object): instance of Blender Object that is wrapped by the child class
            quaternion (RotationParams, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
        """
        self._blender_object = blender_object
        self._tag = tag
        self._quaternion = np.array((1, 0, 0, 0), dtype=np.float64)
        self._translation = np.array((0, 0, 0), dtype=np.float64)
        self.set_position(rotation_mode, rotation, translation)

    @property
    def tag(self):
        return self._tag

    def _rotation_params_to_quat(self, rotation_mode: str = "quaternionWXYZ", rotation: RotationParams = None):
        """
        Converts rotation parameters to quaternion representation
        Args:
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
        Returns:
            np.ndarray: quaternion representation of the rotation
        """
        if rotation is None:
            return None
        if rotation_mode.startswith("quaternion"):
            roll_quat = rotation_mode.lower() == "quaternionxyzw"
            result = np.roll(rotation, 1) if roll_quat else np.array(rotation)
        elif rotation_mode == "rotvec":
            result = np.roll(Rotation.from_rotvec(rotation).as_quat(), 1)
        elif rotation_mode == "rotmat":
            result = np.roll(Rotation.from_matrix(rotation).as_quat(), 1)
        elif rotation_mode.startswith("euler"):
            euler_order = rotation_mode[len("euler"):]
            result = np.roll(Rotation.from_euler(euler_order, rotation, degrees=True).as_quat(), 1)
        elif rotation_mode in ["look_at", "lookat"]:
            look_at_rotation_deg = 0.
            if isinstance(rotation, (tuple, list)) and len(rotation) == 2:
                rotation, look_at_rotation_deg = rotation
            if isinstance(rotation, Positionable):
                look_at = rotation.translation
            else:
                look_at = np.array(rotation)
            translation = np.array(self.translation)
            forward_vec =  translation - look_at
            if (forward_vec_norm:=np.linalg.norm(forward_vec)) < 1e-10:
                result = np.array([1, 0, 0, 0])
            else:
                forward_vec /= forward_vec_norm
                up_vec = np.array([0, 0, 1] if (1 - np.abs(forward_vec[2])) > 1e-10 else [0, 1, 0])
                right_vec = np.cross(up_vec, forward_vec)
                right_vec /= np.linalg.norm(right_vec)
                up_vec = np.cross(forward_vec, right_vec)
                up_vec /= np.linalg.norm(up_vec)
                rotmat = np.stack([right_vec, up_vec, forward_vec], axis=1)
                rot = Rotation.from_matrix(rotmat)
                if np.abs(look_at_rotation_deg) > 1e-10:
                    rot = rot * Rotation.from_euler('Z', look_at_rotation_deg, degrees=True)
                result = np.roll(rot.as_quat(), 1)
        else:
            raise ValueError(f"Unknown rotation mode: {rotation_mode}")
        return result.astype(np.float64)

    def set_position(self, rotation_mode: str = "quaternionWXYZ", rotation: RotationParams = None, translation: Vector3d = None):
        """Sets the position of the object in the scene
        
        Args:
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d): translation vector
        """

        if translation is not None:
            self._translation = np.array(translation, dtype=np.float64)
        if (quaternion := self._rotation_params_to_quat(rotation_mode, rotation)) is not None:
            self._quaternion = quaternion
        self._update_position()

    @property
    def quaternion(self):
        return self._quaternion

    @quaternion.setter
    def quaternion(self, val: Vector4d):
        self._quaternion = np.array(val)
        self._update_position()

    @property
    def translation(self):
        return self._translation

    @translation.setter
    def translation(self, val: Vector3d):
        self._translation = np.array(val)
        self._update_position()

    def _update_position(self):
        self._set_blender_object_position(self._blender_object)

    def _set_blender_object_position(self, blender_object: BlenderGroup):
        def set_position(obj):
            obj.rotation_mode = 'QUATERNION'
            obj.rotation_quaternion = self.quaternion.tolist()
            obj.location = self.translation.tolist()

        if isinstance(blender_object, bpy.types.Collection):
            for obj in blender_object.all_objects.values():
                set_position(obj)
        else:
            set_position(blender_object)

    def _blender_remove_object(self):
        """Removes the object from Blender scene"""
        bpy.ops.object.select_all(action='DESELECT')
        is_collection = isinstance(self._blender_object, bpy.types.Collection)
        if is_collection:
            for obj in self._blender_object.all_objects.values():
                obj.select_set(True)
        else:
            bpy.context.view_layer.objects.active = self._blender_object
            bpy.ops.object.mode_set(mode='OBJECT')
            self._blender_object.select_set(True)
        bpy.ops.object.delete()
        if is_collection:
            bpy.data.collections.remove(self._blender_object)
        self._blender_object = None



"""
Code taken from https://github.com/vguzov/cloudrender/blob/main/cloudrender/camera/trajectory.py
"""
from typing import Sequence

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d
from scipy.signal.windows import gaussian
from scipy.spatial.transform import Rotation, Slerp


class Trajectory:
    def __init__(self):
        self.trajectory = []

    def set_trajectory(self, keypoints):
        self.trajectory = [{k: np.array(v) for k, v in x.items()} for x in keypoints]

    def find_closest_kp_in_traj(self, time: float):
        times = np.array([x['time'] for x in self.trajectory])
        times_diff = times - time
        times_mask = times_diff > 0
        if times_mask.sum() == 0:
            return self.trajectory[-1], self.trajectory[-1]
        times_inds = np.flatnonzero(times_mask)
        curr_ind = times_inds[0]
        if curr_ind == 0:
            return self.trajectory[0], self.trajectory[0]
        return self.trajectory[curr_ind], self.trajectory[curr_ind-1]

    def serialize_trajectory(self):
        """Make trajectory json-serializable

        Returns:
            List[dict]: trajectory keypoints - each keypoint has "time", "position" and "quaternion"
        """
        s_traj = [{k: v.tolist() if isinstance(v, np.ndarray) else v for k, v in x.items()} for x in self.trajectory]
        return s_traj

    def sort_trajectory(self):
        times = [x['time'] for x in self.trajectory]
        ind_sorted = np.argsort(times)
        self.trajectory = [self.trajectory[k] for k in ind_sorted]
        return ind_sorted

    def add_keypoint(self, quaternion: Sequence[float], position: Sequence[float],
                     time: float, check_time: bool = True):
        keypoint = {"position": np.array(position), "quaternion": np.array(quaternion), "time": time}
        if check_time:
            times = np.array([x['time'] for i, x in enumerate(self.trajectory)])
            diff = times - keypoint['time']
            if np.any(np.abs(diff) < 1e-4):
                return
        self.trajectory.append(keypoint)
        self.sort_trajectory()

    def rot_gaussian_smoothing(self, rots, sigma=5.):
        def get_rot_ind(ind):
            while ind >= len(rots) or ind < 0:
                if ind >= len(rots):
                    ind = 2 * len(rots) - 1 - ind
                if ind < 0:
                    ind = -ind
            return ind

        winradius = round(2 * 3 * sigma)
        if winradius < 1:
            return rots
        weights = gaussian(winradius * 2 + 1, sigma)
        res = []
        for ind in range(len(rots)):
            window_inds = [get_rot_ind(i) for i in range(ind - winradius, ind + winradius + 1)]
            res.append(rots[window_inds].mean(weights))
        return res

    def refine_trajectory(self, time_step: float = 1/60., interp_type: str = "quadratic", smoothness: float = 5.0):
        """Refines the trajectory by creating keypoints inbetween existion ones via interpolation

        Args:
            time_step (float): how often to create new points
            interp_type (str): interpolation type, "linear", "quadratic", "cubic"
            smoothness (float): how hard to smooth the pose trajectory

        Returns:
             List[dict]: trajectory keypoints - each keypoint has "time", "position" and "quaternion"
        """
        min_pts_for_interp = {"linear": 2, "quadratic": 3, "cubic": 4}
        assert interp_type in min_pts_for_interp.keys(), \
            f"Available interpolations are: {list(min_pts_for_interp.keys())}"
        if len(self.trajectory) < min_pts_for_interp[interp_type]:
            print(f'Not enough points for interpolation with "{interp_type}", returning unchanged')
            return
        start_time = self.trajectory[0]['time']
        end_time = self.trajectory[-1]['time']
        cam_times = [x['time'] for x in self.trajectory]
        cam_rots = Rotation.from_quat([np.roll(x['quaternion'], -1) for x in self.trajectory])
        cam_poses = [x['position'] for x in self.trajectory]
        rot_slerp = Slerp(cam_times, cam_rots)
        interp_times = np.concatenate([np.arange(start_time, end_time, time_step), [end_time]])
        interp_rots = rot_slerp(interp_times)
        pos_intrp = interp1d(cam_times, cam_poses, axis=0, kind=interp_type)
        interp_poses = pos_intrp(interp_times)
        interp_poses = np.array(list(zip(*[gaussian_filter1d(x, smoothness) for x in zip(*interp_poses)])))
        interp_rots = self.rot_gaussian_smoothing(interp_rots, smoothness)
        interp_quats = [np.roll(x.as_quat(), 1) for x in interp_rots]
        interp_traj = [{'position': interp_poses[i], 'quaternion': interp_quats[i], 'time': interp_times[i]}
                       for i in range(len(interp_times))]
        self.trajectory = interp_traj

        return interp_traj



from abc import abstractmethod
from typing import Optional, List, Union

import bpy
import warnings

from ..colors import UniformColors, VertexColors
from ..colors.base import ColorsMetadata, ColorsList, Colors
from ..colors.texture import TextureColors, FileTextureColors, VideoFileTextureColors, TextureColorsViaTempFile
from ..internal.positionable import Positionable
from ..materials.base import MaterialList, MaterialInstance, Material


class Renderable(Positionable):
    """
    Base class for all renderable objects (Meshes, PointClouds, Primitives).
    """

    @abstractmethod
    def __init__(
            self,
            material: Union[Material, MaterialList],
            colors: Union[Colors, ColorsList],
            **kwargs
    ):
        """Creates internal structures, calls functions that connect Material and Colors to the object.
        Can only be called from child classes as the class is abstract.

        Args:
            material (Union[Material, MaterialList]): Material instance or a list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or a list of Colors instances
            blender_object (bpy.types.Object): instance of Blender Object that is wrapped by the class
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        self._materials_count = len(material) if not isinstance(material, Material) else 1
        self.update_material(material)
        self.update_colors(colors)

    def update_material(self, material: Union[Material, MaterialList]):
        """Updates object material properties, sets Blender structures accordingly
        Args:
            material (Union[Material, MaterialList]): target material or a list of target materials
        """
        pass

    def update_colors(self, colors: Union[Colors, ColorsList]):
        """Updates object color properties, sets Blender structures accordingly

        Args:
            colors (Union[Colors, ColorsList]): target colors information or a list of target colors information
        """
        pass


class RenderableObject(Renderable):
    """
    Base class for renderable objects, that can be represented by a single bpy.types.Object (Meshes and Primitives).
    """

    @abstractmethod
    def __init__(
            self,
            **kwargs
    ):
        """Sets initial values for internal parameters, can only be called from child classes as the class is abstract.

        Args:
            material (MaterialList): Material instance or a list of Material instances
            colors (ColorsList): Colors instance or a list of Colors instances
            blender_object (bpy.types.Object): instance of Blender Object that is wrapped by the class
            quaternion (Vector4d, optional): rotation applied to the Blender object (default: None (identity))
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        self._material_instances: Optional[List[MaterialInstance]] = None
        self._colors_metadatas: Optional[List[ColorsMetadata]] = None
        super().__init__(**kwargs)

    @property
    def emit_shadows(self) -> bool:
        return self._blender_object.visible_shadow

    @emit_shadows.setter
    def emit_shadows(self, val: bool):
        self._blender_object.visible_shadow = val

    # ===================================================== OBJECT =====================================================
    @abstractmethod
    def _blender_create_object(self, *args, **kwargs):
        pass

    def _blender_remove_object(self):
        """Removes the object from Blender scene
        """
        self._blender_clear_colors()
        self._blender_clear_materials()
        super()._blender_remove_object()

    # ================================================== END OF OBJECT =================================================

    # ==================================================== MATERIAL ====================================================
    def update_material(self, material: Union[Material, MaterialList]):
        """Updates object material properties, sets Blender structures accordingly

        Args:
            material (Union[Material, MaterialList]): target material or a list of target materials
        """
        if self._material_instances is not None:
            self._blender_clear_materials()

        # Turn single material into a list
        if isinstance(material, Material):
            material_list = [material]
        else:
            material_list = material

        self._blender_set_materials(material_list)

    def _blender_assign_materials(self):
        """Assigns created materials to the object"""
        blender_material = self._material_instances[0].blender_material
        self._blender_object.active_material = blender_material

    def _blender_set_materials(self, material_list: MaterialList):
        """Constructs material node, recreates color node if needed

        Args:
            material_list (MaterialList): list of target materials
        """
        assert len(material_list) > 0, "Material list must contain at least one material"
        assert len(material_list) == self._materials_count, "Material list must contain the same number of elements as during object creation"

        self._material_instances = []

        for material in material_list:
            material_instance = material.create_material()
            self._blender_object.data.materials.append(material_instance.blender_material)
            self._material_instances.append(material_instance)
        self._blender_create_colors_nodes()
        self._blender_link_color2material()
        self._blender_assign_materials()

    def _blender_clear_materials(self):
        """Clears Blender material node and nodes connected to it
        """
        if self._material_instances is not None:
            self._blender_object.active_material = None
            for material_instance in self._material_instances:
                material_nodes = material_instance.blender_material.node_tree.nodes
                blender_material = material_instance.blender_material
                for material_node in material_nodes:
                    blender_material.node_tree.nodes.remove(material_node)
                blender_material.user_clear()
                bpy.data.materials.remove(blender_material)
            self._material_instances = None

    # ================================================ END OF MATERIAL =================================================

    # ===================================================== COLORS =====================================================
    def update_colors(self, colors: Union[Colors, ColorsList]):
        """Updates object color properties, sets Blender structures accordingly

        Args:
            colors (Colors): target colors information
        """
        if self._material_instances is not None:
            self._blender_clear_colors()

        # Turn single colors into a list
        if isinstance(colors, Colors):
            colors_list = [colors]
        else:
            colors_list = colors
        self._blender_set_colors(colors_list)

    def _blender_set_colors(self, colors_list: ColorsList):
        """Remembers current color properties, builds a color node for material (from colors_metadata)
        """
        assert len(colors_list) > 0, "Colors list must contain at least one color"
        assert len(colors_list) == self._materials_count, "Colors list must contain the same number of elements as materials"

        self._colors_metadatas = []
        for material_ind, colors in enumerate(colors_list):
            self._colors_metadatas.append(colors.metadata)
        self._blender_create_colors_nodes()
        self._blender_link_color2material()

    def _blender_clear_colors(self):
        """Clears Blender color node and erases node constructor
        """
        if self._material_instances is not None and self._colors_metadatas is not None:
            for material_instance, colors_metadata in zip(self._material_instances, self._colors_metadatas):
                if colors_metadata.type is UniformColors and colors_metadata.has_alpha:
                    # return the alpha channel back to the default value
                    material_instance.inputs['Alpha'].default_value = 1.
                blender_material = material_instance.blender_material
                blender_material.node_tree.nodes.remove(material_instance.colors_node)
                material_instance.colors_node = None
        self._colors_metadatas = None

    def _blender_create_colors_nodes(self):
        """Creates color node using previously set builder
        """
        if self._colors_metadatas is not None and self._material_instances is not None:
            for material_instance, colors_metadata in zip(self._material_instances, self._colors_metadatas):
                blender_material = material_instance.blender_material
                if colors_metadata.type is UniformColors:
                    colors_node = blender_material.node_tree.nodes.new('ShaderNodeRGB')
                    if colors_metadata.has_alpha:
                        colors_node.outputs["Color"].default_value = colors_metadata.color.tolist()
                    else:
                        colors_node.outputs["Color"].default_value = colors_metadata.color.tolist() + [1.]  # add alpha 1.0
                elif colors_metadata.type is VertexColors:
                    colors_node = blender_material.node_tree.nodes.new('ShaderNodeVertexColor')
                elif colors_metadata.type in (TextureColors, FileTextureColors, TextureColorsViaTempFile, VideoFileTextureColors):
                    colors_node = blender_material.node_tree.nodes.new('ShaderNodeTexImage')
                    colors_node.image = colors_metadata.texture
                    colors_node.interpolation = colors_metadata.interpolation
                    if colors_metadata.type is VideoFileTextureColors:
                        colors_node.image_user.frame_duration = colors_node.image.frame_duration
                        colors_node.image_user.frame_start = colors_metadata.frame_start + 1
                else:
                    raise NotImplementedError(f"Unsupported colors class '{colors_metadata.type}'")
                material_instance.colors_node = colors_node

    def _blender_link_color2material(self):
        """Links color and material nodes
        """
        if self._material_instances is not None and self._colors_metadatas is not None:
            for material_instance, colors_metadata in zip(self._material_instances, self._colors_metadatas):
                if material_instance.colors_node is not None:
                    blender_material = material_instance.blender_material
                    blender_material.node_tree.links.new(material_instance.inputs['Color'],
                                                         material_instance.colors_node.outputs['Color'])
                    material_instance.inputs['Color'].default_value = [1.0, 0.0, 0.0, 1.0]
                    if colors_metadata.has_alpha:
                        if 'Alpha' in material_instance.inputs:
                            if colors_metadata.type is UniformColors:
                                material_instance.inputs['Alpha'].default_value = colors_metadata.color[3]
                            else:
                                blender_material.node_tree.links.new(material_instance.inputs['Alpha'],
                                                                     material_instance.colors_node.outputs['Alpha'])
                        else:
                            warnings.warn("This material does not support transparency; alpha color channel is ignored")
    # ================================================== END OF COLORS =================================================
